from __future__ import annotations

from typing import Literal

from prettyqt import constants, core, widgets
from prettyqt.qt import QtCore, QtWidgets
from prettyqt.utils import InvalidParamError, bidict


SLIDER_ACTION = bidict(
    none=QtWidgets.QAbstractSlider.SliderNoAction,
    single_step_add=QtWidgets.QAbstractSlider.SliderSingleStepAdd,
    single_step_sub=QtWidgets.QAbstractSlider.SliderSingleStepSub,
    page_step_add=QtWidgets.QAbstractSlider.SliderPageStepAdd,
    page_step_sub=QtWidgets.QAbstractSlider.SliderPageStepSub,
    to_minimum=QtWidgets.QAbstractSlider.SliderToMinimum,
    to_maximum=QtWidgets.QAbstractSlider.SliderToMaximum,
    move=QtWidgets.QAbstractSlider.SliderMove,
)

SliderActionStr = Literal[
    "none",
    "single_step_add",
    "single_step_sub",
    "page_step_add",
    "page_step_sub",
    "to_minimum",
    "to_maximum",
    "move",
]

QtWidgets.QAbstractSlider.__bases__ = (widgets.Widget,)


class AbstractSlider(QtWidgets.QAbstractSlider):

    value_changed = core.Signal(int)

    def on_value_change(self):
        self.value_changed.emit(self.value())

    def serialize_fields(self):
        return dict(
            range=(self.minimum(), self.maximum()),
            value=self.value(),
            has_tracking=self.hasTracking(),
            inverted_controls=self.invertedControls(),
            inverted_appearance=self.invertedAppearance(),
            single_step=self.singleStep(),
            page_step=self.pageStep(),
        )

    def __setstate__(self, state):
        super().__setstate__(state)
        self.set_range(*state["range"])
        self.set_value(state["value"])
        self.setSingleStep(state["single_step"])
        self.setPageStep(state["page_step"])
        self.setTracking(state["has_tracking"])
        self.setInvertedControls(state["inverted_controls"])
        self.setInvertedAppearance(state["inverted_appearance"])

    def __reduce__(self):
        return type(self), (), self.__getstate__()

    def is_horizontal(self) -> bool:
        """Check if silder is horizontal.

        Returns:
            True if horizontal, else False
        """
        return self.orientation() == QtCore.Qt.Horizontal

    def is_vertical(self) -> bool:
        """Check if silder is vertical.

        Returns:
            True if vertical, else False
        """
        return self.orientation() == QtCore.Qt.Vertical

    def set_horizontal(self):
        """Set slider orientation to horizontal."""
        self.setOrientation(QtCore.Qt.Horizontal)

    def set_vertical(self):
        """Set slider orientation to vertical."""
        self.setOrientation(QtCore.Qt.Vertical)

    def set_orientation(self, orientation: constants.OrientationStr):
        """Set the orientation of the slider.

        Args:
            orientation: orientation for the slider

        Raises:
            InvalidParamError: orientation does not exist
        """
        if orientation not in constants.ORIENTATION:
            raise InvalidParamError(orientation, constants.ORIENTATION)
        self.setOrientation(constants.ORIENTATION[orientation])

    def get_orientation(self) -> constants.OrientationStr:
        """Return current orientation.

        Returns:
            orientation
        """
        return constants.ORIENTATION.inverse[self.orientation()]

    def scroll_to_min(self):
        """Scroll to the minimum value of the slider."""
        self.setValue(self.minimum())

    def scroll_to_max(self):
        """Scroll to the maximum value of the slider."""
        self.setValue(self.maximum())

    def set_range(self, min_val: int, max_val: int):
        self.setRange(min_val, max_val)

    def set_step_size(self, step_size: int):
        self.setSingleStep(step_size)

    def set_repeat_action(
        self, action: SliderActionStr, threshold: int = 500, repeat_time: int = 50
    ):
        """Set the repeat action.

        Args:
            action: repeat action

        Raises:
            InvalidParamError: invalid repeat action
        """
        if action not in SLIDER_ACTION:
            raise InvalidParamError(action, SLIDER_ACTION)
        self.setRepeatAction(SLIDER_ACTION[action], threshold, repeat_time)

    def get_repeat_action(self) -> SliderActionStr:
        """Get current repeat action.

        Returns:
            current repeat action
        """
        return SLIDER_ACTION.inverse[self.repeatAction()]

    def trigger_action(self, action: SliderActionStr):
        """Trigger slider action."""
        if action not in SLIDER_ACTION:
            raise InvalidParamError(action, SLIDER_ACTION)
        self.triggerAction(SLIDER_ACTION[action])

    def get_value(self):
        return self.value()

    def set_value(self, value: int):
        self.setValue(value)


if __name__ == "__main__":
    app = QtWidgets.QApplication([])
    slider = AbstractSlider()
    slider.setRange(0, 100)
    slider.value_changed.connect(print)
    slider.show()
    app.main_loop()



from __future__ import annotations

import contextlib
from typing import Iterator, Literal

from prettyqt import core, gui, widgets
from prettyqt.qt import QtWidgets
from prettyqt.utils import InvalidParamError, bidict, colors, mappers, types


AUTO_FORMATTING = mappers.FlagMap(
    QtWidgets.QTextEdit.AutoFormatting,
    none=QtWidgets.QTextEdit.AutoNone,
    bullet_list=QtWidgets.QTextEdit.AutoBulletList,
    all=QtWidgets.QTextEdit.AutoAll,
)

AutoFormattingStr = Literal["none", "bullet_list", "all"]

LINE_WRAP_MODE = bidict(
    none=QtWidgets.QTextEdit.NoWrap,
    widget_width=QtWidgets.QTextEdit.WidgetWidth,
    fixed_pixel_width=QtWidgets.QTextEdit.FixedPixelWidth,
    fixed_column_width=QtWidgets.QTextEdit.FixedColumnWidth,
)

LineWrapModeStr = Literal[
    "none", "widget_width", "fixed_pixel_width", "fixed_column_width"
]


QtWidgets.QTextEdit.__bases__ = (widgets.AbstractScrollArea,)


class TextEdit(QtWidgets.QTextEdit):

    value_changed = core.Signal(str)

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.textChanged.connect(self.on_value_change)

    def serialize_fields(self):
        return dict(
            text=self.text(),
            accept_rich_text=self.acceptRichText(),
            auto_formatting=self.get_auto_formatting(),
            cursor_width=self.cursorWidth(),
            document_title=self.documentTitle(),
            line_wrap_column_or_width=self.lineWrapColumnOrWidth(),
            line_wrap_mode=self.get_line_wrap_mode(),
            word_wrap_mode=self.get_word_wrap_mode(),
            overwrite_mode=self.overwriteMode(),
            placeholder_text=self.placeholderText(),
            read_only=self.isReadOnly(),
            tab_changes_focus=self.tabChangesFocus(),
            tab_stop_distance=self.tabStopDistance(),
            undo_redo_enabled=self.isUndoRedoEnabled(),
        )

    def __setstate__(self, state):
        super().__setstate__(state)
        self.set_text(state["text"])
        self.setAcceptRichText(state["accept_rich_text"])
        self.set_auto_formatting(state["auto_formatting"])
        self.setCursorWidth(state["cursor_width"])
        self.setDocumentTitle(state["document_title"])
        self.setLineWrapColumnOrWidth(state["line_wrap_column_or_width"])
        self.set_line_wrap_mode(state["line_wrap_mode"])
        self.set_word_wrap_mode(state["word_wrap_mode"])
        self.setOverwriteMode(state["overwrite_mode"])
        self.setPlaceholderText(state["placeholder_text"])
        self.setReadOnly(state["read_only"])
        self.setTabChangesFocus(state["tab_changes_focus"])
        self.setTabStopDistance(state["tab_stop_distance"])
        self.setUndoRedoEnabled(state["undo_redo_enabled"])

    def __reduce__(self):
        return type(self), (), self.__getstate__()

    def __add__(self, other: str) -> TextEdit:
        self.append_text(other)
        return self

    def on_value_change(self) -> None:
        self.value_changed.emit(self.text())

    @contextlib.contextmanager
    def create_cursor(self) -> Iterator[gui.TextCursor]:
        cursor = gui.TextCursor(self.document())
        yield cursor
        self.setTextCursor(cursor)

    def set_text(self, text: str) -> None:
        self.setPlainText(text)

    def append_text(self, text: str) -> None:
        self.append(text)

    def text(self) -> str:
        return self.toPlainText()

    def select_text(self, start: int, end: int) -> None:
        with self.create_cursor() as c:
            c.select_text(start, end)

    def set_read_only(self, value: bool = True) -> None:
        self.setReadOnly(value)

    def set_text_color(self, color: types.ColorType) -> None:
        color = colors.get_color(color)
        self.setTextColor(color)

    def set_line_wrap_mode(self, mode: LineWrapModeStr):
        """Set line wrap mode.

        Args:
            mode: line wrap mode to use

        Raises:
            InvalidParamError: line wrap mode does not exist
        """
        if mode not in LINE_WRAP_MODE:
            raise InvalidParamError(mode, LINE_WRAP_MODE)
        self.setLineWrapMode(LINE_WRAP_MODE[mode])

    def get_line_wrap_mode(self) -> LineWrapModeStr:
        """Get the current wrap mode.

        Returns:
            Wrap mode
        """
        return LINE_WRAP_MODE.inverse[self.lineWrapMode()]

    def set_auto_formatting(self, mode: AutoFormattingStr):
        """Set auto formatting mode.

        Args:
            mode: auto formatting mode to use

        Raises:
            InvalidParamError: auto formatting mode does not exist
        """
        if mode not in AUTO_FORMATTING:
            raise InvalidParamError(mode, AUTO_FORMATTING)
        self.setAutoFormatting(AUTO_FORMATTING[mode])

    def get_auto_formatting(self) -> AutoFormattingStr:
        """Get the current auto formatting mode.

        Returns:
            Auto formatting mode
        """
        return AUTO_FORMATTING.inverse[self.autoFormatting()]

    def set_word_wrap_mode(self, mode: gui.textoption.WordWrapModeStr):
        """Set word wrap mode.

        Args:
            mode: word wrap mode to use

        Raises:
            InvalidParamError: wrap mode does not exist
        """
        if mode not in gui.textoption.WORD_WRAP_MODE:
            raise InvalidParamError(mode, gui.textoption.WORD_WRAP_MODE)
        self.setWordWrapMode(gui.textoption.WORD_WRAP_MODE[mode])

    def get_word_wrap_mode(self) -> gui.textoption.WordWrapModeStr:
        """Get the current word wrap mode.

        Returns:
            Word wrap mode
        """
        return gui.textoption.WORD_WRAP_MODE.inverse[self.wordWrapMode()]


if __name__ == "__main__":
    app = QtWidgets.QApplication([])
    widget = TextEdit("This is a test")
    widget.show()
    app.main_loop()



from __future__ import annotations

from typing import Literal, Optional

from prettyqt import core
from prettyqt.qt import QtNetwork
from prettyqt.utils import InvalidParamError, bidict


LOCAL_SOCKET_ERROR = bidict(
    connection_refused=QtNetwork.QLocalSocket.ConnectionRefusedError,
    peer_closed=QtNetwork.QLocalSocket.PeerClosedError,
    server_not_found=QtNetwork.QLocalSocket.ServerNotFoundError,
    socket_access=QtNetwork.QLocalSocket.SocketAccessError,
    socket_resource=QtNetwork.QLocalSocket.SocketResourceError,
    socket_timeout=QtNetwork.QLocalSocket.SocketTimeoutError,
    datagram_too_large=QtNetwork.QLocalSocket.DatagramTooLargeError,
    connection=QtNetwork.QLocalSocket.ConnectionError,
    unsupported_socket_operation=QtNetwork.QLocalSocket.UnsupportedSocketOperationError,
    operation=QtNetwork.QLocalSocket.OperationError,
    unknown_socket=QtNetwork.QLocalSocket.UnknownSocketError,
)

LocalSocketErrorStr = Literal[
    "connection_refused",
    "peer_closed",
    "server_not_found",
    "socket_access",
    "socket_resource",
    "socket_timeout",
    "datagram_too_large",
    "connection",
    "unsupported_socket_operation",
    "operation",
    "unknown_socket",
]

LOCAL_SOCKET_STATE = bidict(
    unconnected=QtNetwork.QLocalSocket.UnconnectedState,
    connecting=QtNetwork.QLocalSocket.ConnectingState,
    connected=QtNetwork.QLocalSocket.ConnectedState,
    closing=QtNetwork.QLocalSocket.ClosingState,
)

LocalSocketStateStr = Literal[
    "unconnected",
    "connecting",
    "connected",
    "closing",
]


QtNetwork.QLocalSocket.__bases__ = (core.IODevice,)


class LocalSocket(QtNetwork.QLocalSocket):
    def __bool__(self):
        return self.isValid()

    def get_error(self) -> LocalSocketErrorStr:
        return LOCAL_SOCKET_ERROR.inverse[self.error()]

    def get_state(self) -> LocalSocketStateStr:
        return LOCAL_SOCKET_STATE.inverse[self.state()]

    def connect_to_server(
        self,
        server_name: Optional[str] = None,
        mode: core.iodevice.OpenModeStr = "read_write",
    ):
        if mode not in core.iodevice.OPEN_MODES:
            raise InvalidParamError(mode, core.iodevice.OPEN_MODES)
        if server_name is not None:
            self.connectToServer(server_name, core.iodevice.OPEN_MODES[mode])
        else:
            self.connectToServer(core.iodevice.OPEN_MODES[mode])


if __name__ == "__main__":
    server = LocalSocket()



from __future__ import annotations

from typing import Literal, Optional, Tuple, Union

from deprecated import deprecated

from prettyqt import constants, core, widgets
from prettyqt.qt import QtCore, QtGui, QtWidgets
from prettyqt.utils import InvalidParamError, bidict


REMOVE_BEHAVIOUR = bidict(
    left_tab=QtWidgets.QTabBar.SelectLeftTab,
    right_tab=QtWidgets.QTabBar.SelectRightTab,
    previous_tab=QtWidgets.QTabBar.SelectPreviousTab,
)

RemoveBehaviourStr = Literal["left_tab", "right_tab", "previous_tab"]

SHAPE = bidict(
    rounded_north=QtWidgets.QTabBar.RoundedNorth,
    rounded_south=QtWidgets.QTabBar.RoundedSouth,
    rounded_west=QtWidgets.QTabBar.RoundedWest,
    rounded_east=QtWidgets.QTabBar.RoundedEast,
    triangular_north=QtWidgets.QTabBar.TriangularNorth,
    triangular_south=QtWidgets.QTabBar.TriangularSouth,
    triangular_west=QtWidgets.QTabBar.TriangularWest,
    triangular_east=QtWidgets.QTabBar.TriangularEast,
)

ShapeStr = Literal[
    "rounded_north",
    "rounded_south",
    "rounded_west",
    "rounded_east",
    "triangular_north",
    "triangular_south",
    "triangular_west",
    "triangular_east",
]

POSITIONS = bidict(left=QtWidgets.QTabBar.LeftSide, right=QtWidgets.QTabBar.RightSide)

PositionStr = Literal["left", "right"]

QtWidgets.QTabBar.__bases__ = (widgets.Widget,)


class TabBar(QtWidgets.QTabBar):
    on_detach = QtCore.Signal(int, QtCore.QPoint)

    def __init__(self, parent: Optional[QtWidgets.QWidget] = None) -> None:
        super().__init__(parent)

        self.setAcceptDrops(True)
        self.set_elide_mode("right")
        self.set_selection_behavior_on_remove("left_tab")

    def __getitem__(self, index: Tuple[int, str]):
        return self.tabButton(index[0], POSITIONS[index[1]])

    def __setitem__(
        self, index: Tuple[int, PositionStr], value: Optional[QtWidgets.QWidget]
    ):
        self.set_tab(index[0], index[1], value)

    def serialize_fields(self):
        return dict(
            movable=self.isMovable(),
            document_mode=self.documentMode(),
            current_index=self.currentIndex(),
            # shape=self.shape(),
            draw_base=self.drawBase(),
            elide_mode=self.get_elide_mode(),
            icon_size=core.Size(self.iconSize()),
        )

    def __setstate__(self, state):
        self.setDocumentMode(state.get("document_mode", False))
        self.setMovable(state.get("movable", False))
        # self.setShape(state.get("shape", "rounded"))
        self.setIconSize(state["icon_size"])
        self.setDrawBase(state.get("draw_base"))
        self.set_elide_mode(state.get("elide_mode"))
        self.setCurrentIndex(state.get("index", 0))

    def __reduce__(self):
        return type(self), (), self.__getstate__()

    #  Send the on_detach when a tab is double clicked
    def mouseDoubleClickEvent(self, event):
        event.accept()
        tab = self.tabAt(event.pos())
        pos = QtGui.QCursor.pos()
        self.on_detach.emit(tab, pos)

    def set_icon_size(self, size: Union[int, Tuple[int, int], QtCore.QSize]):
        """Set size of the icons."""
        if isinstance(size, int):
            size = core.Size(size, size)
        elif isinstance(size, tuple):
            size = core.Size(*size)
        self.setIconSize(size)

    def get_icon_size(self) -> core.Size:
        return core.Size(self.iconSize())

    def set_tab(
        self, index: int, position: PositionStr, widget: Optional[QtWidgets.QWidget]
    ) -> None:
        self.setTabButton(index, POSITIONS[position], widget)  # type: ignore

    @deprecated(
        reason="This method is deprecated, use set_selection_behavior_on_remove instead."
    )
    def set_remove_behaviour(self, mode: RemoveBehaviourStr) -> None:
        self.set_selection_behavior_on_remove(mode)

    def set_selection_behavior_on_remove(self, mode: RemoveBehaviourStr) -> None:
        """Set the remove hehaviour.

        What tab should be set as current when removeTab is called
        if the removed tab is also the current tab.

        Args:
            mode: new remove behaviour
        """
        if mode not in REMOVE_BEHAVIOUR:
            raise InvalidParamError(mode, REMOVE_BEHAVIOUR)
        self.setSelectionBehaviorOnRemove(REMOVE_BEHAVIOUR[mode])

    def get_remove_behaviour(self) -> RemoveBehaviourStr:
        """Return remove behaviour.

        Returns:
            remove behaviour
        """
        return REMOVE_BEHAVIOUR.inverse[self.selectionBehaviorOnRemove()]

    def set_elide_mode(self, mode: constants.ElideModeStr) -> None:
        """Set elide mode.

        Args:
            mode: elide mode to use

        Raises:
            InvalidParamError: invalid elide mode
        """
        if mode not in constants.ELIDE_MODE:
            raise InvalidParamError(mode, constants.ELIDE_MODE)
        self.setElideMode(constants.ELIDE_MODE[mode])

    def get_elide_mode(self) -> constants.ElideModeStr:
        """Return elide mode.

        Returns:
            elide mode
        """
        return constants.ELIDE_MODE.inverse[self.elideMode()]



from __future__ import annotations

from typing import Iterator

from prettyqt import widgets
from prettyqt.qt import QtWidgets


QtWidgets.QStackedWidget.__bases__ = (widgets.Frame,)


class StackedWidget(QtWidgets.QStackedWidget):
    def __setstate__(self, state):
        for item in state["items"]:
            self.add(item)

    def __reduce__(self):
        return type(self), (), self.__getstate__()

    def __add__(self, other: QtWidgets.QWidget) -> StackedWidget:
        self.addWidget(other)
        return self

    def __getitem__(self, index: int) -> QtWidgets.QWidget:
        return self.widget(index)

    def __iter__(self) -> Iterator[QtWidgets.QWidget]:
        return iter(self.widget(i) for i in range(self.count()))

    def __len__(self):
        # needed for PySide2
        return self.count()

    def __contains__(self, item: QtWidgets.QWidget):
        return self.indexOf(item) >= 0

    def serialize_fields(self):
        return dict(items=list(self))

    def set_current_widget(self, widget: QtWidgets.QWidget):
        self.setCurrentWidget(widget)


if __name__ == "__main__":
    from prettyqt import widgets

    app = widgets.app()
    stackedwidget = StackedWidget()
    widget2 = widgets.RadioButton("Test")
    widget3 = widgets.RadioButton("Test 2")
    stackedwidget += widget2
    stackedwidget += widget3
    stackedwidget.show()
    app.main_loop()



from __future__ import annotations

from typing import Union

from prettyqt import core
from prettyqt.qt import QtCore


class JsonDocument(QtCore.QJsonDocument):
    def __str__(self):
        return str(self.toVariant())

    def __repr__(self):
        return f"{type(self).__name__}({self.toVariant()!r})"

    def __getitem__(self, index: Union[int, str]):
        val = self.array() if self.isArray() else self.object()
        return core.JsonValue(val[index])  # type: ignore

    def __setitem__(self, index: Union[int, str], value):
        if self.isArray():
            if not isinstance(index, int):
                raise TypeError()
            array = self.array()
            array[index] = value
            self.setArray(array)
        elif self.isObject():
            if not isinstance(index, str):
                raise TypeError()
            obj = self.object()
            obj[index] = value
            self.setObject(obj)

    def to_string(self, indented: bool = False) -> str:
        flag = QtCore.QJsonDocument.Indented if indented else QtCore.QJsonDocument.Compact
        return bytes(self.toJson(flag)).decode()

    @classmethod
    def from_variant(cls, obj):
        doc = cls.fromVariant(obj)
        new = cls()
        if doc.isArray():
            new.setArray(doc.array())
        else:
            new.setObject(doc.object())
        return new


if __name__ == "__main__":
    doc = JsonDocument.from_variant(dict(a="b"))
    print(doc.toVariant())
    print(str(doc))
    new = JsonDocument()
    new.setObject(doc.object())
    print("test")
    print(type(new["a"]))
    print(str(new["a"]))
    new["k"] = "v"
    print("new", new["k"])
    print(new.to_json())



from __future__ import annotations

from prettyqt import core
from prettyqt.qt import QtCore


QtCore.QSortFilterProxyModel.__bases__ = (core.AbstractProxyModel,)


class SortFilterProxyModel(QtCore.QSortFilterProxyModel):
    def serialize_fields(self):
        return dict(
            dynamic_sort_filter=self.dynamicSortFilter(),
            is_filter_case_sensitive=self.is_filter_case_sensitive(),
            filter_key_column=self.filterKeyColumn(),
            filter_regular_expression=self.get_filter_regular_expression(),
            filter_role=self.filterRole(),
            is_sort_locale_aware=self.isSortLocaleAware(),
            recursive_filtering_enabled=self.isRecursiveFilteringEnabled(),
            is_sort_case_sensitive=self.is_sort_case_sensitive(),
            sort_role=self.sortRole(),
        )

    # def set_filter_case_sensitivity(self, sensitivity: constants.CaseSensitivityStr):
    #     """Set the filter case sensitivity.

    #     Args:
    #         sensitivity: filter case sensitivity

    #     Raises:
    #         InvalidParamError: filter case sensitivity does not exist
    #     """
    #     if sensitivity not in constants.CASE_SENSITIVITY:
    #         raise InvalidParamError(sensitivity, constants.CASE_SENSITIVITY)
    #     self.setFilterCaseSensitivity(constants.CASE_SENSITIVITY[sensitivity])

    # def get_filter_case_sensitivity(self) -> constants.CaseSensitivityStr:
    #     """Return current filter case sensitivity.

    #     Returns:
    #         filter case sensitivity
    #     """
    #     return constants.CASE_SENSITIVITY.inverse[self.filterCaseSensitivity()]

    def set_filter_case_sensitive(self, state: bool):
        if state:
            sensitivity = QtCore.Qt.CaseSensitive
        else:
            sensitivity = QtCore.Qt.CaseInsensitive
        self.setFilterCaseSensitivity(sensitivity)

    def is_filter_case_sensitive(self) -> bool:
        return self.filterCaseSensitivity() == QtCore.Qt.CaseSensitive

    def set_sort_case_sensitive(self, state: bool):
        if state:
            sensitivity = QtCore.Qt.CaseSensitive
        else:
            sensitivity = QtCore.Qt.CaseInsensitive
        self.setSortCaseSensitivity(sensitivity)

    def is_sort_case_sensitive(self) -> bool:
        return self.sortCaseSensitivity() == QtCore.Qt.CaseSensitive

    def get_filter_regular_expression(self) -> core.RegularExpression:
        return core.RegularExpression(self.filterRegularExpression())



from __future__ import annotations

from prettyqt.qt import QtGui


class Matrix4x4(QtGui.QMatrix4x4):
    # def __repr__(self):
    #     return f"{type(self).__name__}()"
    pass


if __name__ == "__main__":
    matrix = Matrix4x4([1, 2, 3, 4, 1, 2, 3, 4, 1, 2, 3, 4, 1, 2, 3, 4])
    print(matrix)



from __future__ import annotations

from prettyqt import widgets
from prettyqt.qt import QtWidgets


QtWidgets.QStyleOptionRubberBand.__bases__ = (widgets.StyleOption,)


class StyleOptionRubberBand(QtWidgets.QStyleOptionRubberBand):
    pass



from __future__ import annotations

import math
from typing import Iterator, List, Tuple, Union

from prettyqt import core, gui
from prettyqt.qt import QtCore, QtGui


class PolygonF(QtGui.QPolygonF):
    def __repr__(self):
        points_str = ", ".join([repr(i) for i in self])
        return f"{type(self).__name__}({points_str})"

    def __iter__(self) -> Iterator[core.PointF]:
        return iter(self.get_point(i) for i in range(self.size()))

    def __len__(self) -> int:
        return self.size()

    def __bool__(self):
        return not self.isEmpty()

    def __contains__(self, point: QtCore.QPointF) -> bool:
        return self.containsPoint(point, QtCore.Qt.OddEvenFill)

    def __getitem__(self, index: int) -> core.PointF:
        if index >= self.size():
            raise KeyError(index)
        return self.get_point(index)

    # def __setitem__(self, index: int, value: Union[QtCore.QPoint, Tuple[int, int]]):
    #     if isinstance(value, tuple):
    #         self.setPoint(index, *value)
    #     else:
    #         self.setPoint(index, value)

    def __sub__(self, other: QtGui.QPolygonF) -> PolygonF:
        return PolygonF(self.subtracted(other))

    def __and__(self, other: QtGui.QPolygonF) -> PolygonF:  # &
        return PolygonF(self.intersected(other))

    def __xor__(self, other: QtGui.QPolygonF) -> PolygonF:  # ^
        union = self | other
        intersect = self & other
        return union - intersect

    def __or__(self, other: QtGui.QPolygonF) -> PolygonF:  # |
        return PolygonF(self.united(other))

    def __reduce__(self):
        return type(self), (), self.__getstate__()

    def __getstate__(self):
        return bytes(self)

    def __setstate__(self, ba):
        core.DataStream.write_bytearray(ba, self)

    def __bytes__(self):
        ba = core.DataStream.create_bytearray(self)
        return bytes(ba)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, type(self)):
            return False
        return list(self) == [other.at(i) for i in range(other.size())]

    def get_point(self, index: int) -> core.PointF:
        return core.PointF(self.at(index))

    def get_points(self) -> List[core.PointF]:
        return [self.get_point(i) for i in range(self.size())]

    def add_points(self, *points: Union[Tuple[float, float], core.Point]):
        for p in points:
            if isinstance(p, tuple):
                p = core.Point(*p)
            self.append(p)

    def to_polygon(self) -> gui.Polygon:
        return gui.Polygon(self.toPolygon())

    @classmethod
    def create_star(cls):
        poly = cls()
        poly.append(core.PointF(1.0, 0.5))
        for i in range(1, 5):
            val = 0.8 * i * math.pi
            point = core.PointF(0.5 + 0.5 * math.cos(val), 0.5 + 0.5 * math.sin(val))
            poly.append(point)
        return poly

    @classmethod
    def create_diamond(cls):
        points = [
            core.PointF(0.4, 0.5),
            core.PointF(0.5, 0.4),
            core.PointF(0.6, 0.5),
            core.PointF(0.5, 0.6),
            core.PointF(0.4, 0.5),
        ]
        poly = gui.PolygonF()
        poly.add_points(*points)
        return poly

    @classmethod
    def from_xy(cls, xdata, ydata):
        import numpy as np

        size = len(xdata)
        polyline = cls(size)
        pointer = polyline.data()
        dtype, tinfo = np.float, np.finfo  # integers: = np.int, np.iinfo
        pointer.setsize(2 * polyline.size() * tinfo(dtype).dtype.itemsize)
        memory = np.frombuffer(pointer, dtype)
        memory[: (size - 1) * 2 + 1 : 2] = xdata
        memory[1 : (size - 1) * 2 + 2 : 2] = ydata
        return polyline


if __name__ == "__main__":
    poly = PolygonF([core.Point(1, 1), core.Point(2, 2)])
    poly2 = PolygonF([core.Point(1, 1), core.Point(2, 2)])
    new = poly | poly2
    print(repr(new))



from __future__ import annotations

from prettyqt import core, widgets
from prettyqt.qt import QtWidgets


QtWidgets.QGraphicsEllipseItem.__bases__ = (widgets.AbstractGraphicsShapeItem,)


class GraphicsEllipseItem(QtWidgets.QGraphicsEllipseItem):
    def __repr__(self):
        return f"{type(self).__name__}({repr(self.get_rect())})"

    def serialize_fields(self):
        return dict(
            rect=self.get_rect(),
            span_angle=self.spanAngle(),
            start_angle=self.startAngle(),
        )

    def get_rect(self) -> core.Rect:
        return core.Rect(self.rect())



from __future__ import annotations

from prettyqt import widgets
from prettyqt.qt import QtWidgets


QtWidgets.QStyleOptionGroupBox.__bases__ = (widgets.StyleOptionComplex,)


class StyleOptionGroupBox(QtWidgets.QStyleOptionGroupBox):
    pass



"""qthelp module.

contains QtHelp-based classes
"""

from .helpindexmodel import HelpIndexModel
from .helpcontentitem import HelpContentItem
from .helplink import HelpLink
from .helpsearchresult import HelpSearchResult
from .helpcontentmodel import HelpContentModel
from .helpcontentwidget import HelpContentWidget
from .helpfilterdata import HelpFilterData
from .helpindexwidget import HelpIndexWidget
from .helpsearchresultwidget import HelpSearchResultWidget
from .helpsearchquerywidget import HelpSearchQueryWidget
from .helpenginecore import HelpEngineCore
from .helpengine import HelpEngine
from .helpsearchengine import HelpSearchEngine

from prettyqt import core

if core.VersionNumber.get_qt_version() >= (5, 13, 0):
    from .helpfilterengine import HelpFilterEngine

__all__ = [
    "HelpIndexModel",
    "HelpContentItem",
    "HelpLink",
    "HelpSearchResult",
    "HelpContentModel",
    "HelpContentWidget",
    "HelpFilterData",
    "HelpIndexWidget",
    "HelpSearchEngine",
    "HelpSearchResultWidget",
    "HelpSearchQueryWidget",
    "HelpFilterEngine",
    "HelpEngineCore",
    "HelpEngine",
]



from __future__ import annotations

import logging
import os
import pathlib
import sys
from typing import Callable, Dict, List, Optional, Tuple, Union

from prettyqt import constants, core
from prettyqt.qt import QtCore
from prettyqt.utils import InvalidParamError


logger = logging.getLogger(__name__)

QtCore.QCoreApplication.__bases__ = (core.Object,)


class CoreApplication(QtCore.QCoreApplication):
    translators: Dict[str, core.Translator] = dict()

    @classmethod
    def call_on_exit(cls, func: Callable):
        cls.instance().aboutToQuit.connect(func)

    @classmethod
    def get_application_file_path(cls) -> pathlib.Path:
        return pathlib.Path(cls.applicationFilePath())

    @classmethod
    def get_application_dir_path(cls) -> pathlib.Path:
        return pathlib.Path(cls.applicationDirPath())

    @classmethod
    def add_library_path(cls, path: Union[os.PathLike, str]):
        cls.addLibraryPath(os.fspath(path))

    @classmethod
    def get_library_paths(cls) -> List[pathlib.Path]:
        return [pathlib.Path(i) for i in cls.libraryPaths()]

    @classmethod
    def use_hdpi_bitmaps(cls, state: bool = True):
        cls.setAttribute(QtCore.Qt.AA_UseHighDpiPixmaps, state)

    @classmethod
    def disable_window_help_button(cls, state: bool = True):
        try:
            cls.setAttribute(QtCore.Qt.AA_DisableWindowContextHelpButton, state)
        except AttributeError:
            pass

    def set_metadata(
        self,
        app_name: Optional[str] = None,
        app_version: Union[None, str, QtCore.QVersionNumber, Tuple[int, int, int]] = None,
        org_name: Optional[str] = None,
        org_domain: Optional[str] = None,
    ):
        if app_name is not None:
            self.setApplicationName(app_name)
        if app_version is not None:
            if isinstance(app_version, QtCore.QVersionNumber):
                app_version = app_version.toString()
            elif isinstance(app_version, tuple):
                app_version = ".".join(str(i) for i in app_version)
            self.setApplicationVersion(app_version)
        if org_name is not None:
            self.setOrganizationName(org_name)
        if org_domain is not None:
            self.setOrganizationDomain(org_domain)
        # if sys.platform.startswith("win"):
        #     import ctypes
        #     myappid = 'mycompany.myproduct.subproduct.version' # arbitrary string
        #     ctypes.windll.shell32.SetCurrentProcessExplicitAppUserModelID(myappid)

    @classmethod
    def load_language_file(cls, file: Union[pathlib.Path, str]) -> core.Translator:
        translator = core.Translator()
        translator.load_file(file)
        cls.installTranslator(translator)
        cls.translators[str(file)] = translator
        return translator

    @classmethod
    def load_language(cls, language: str) -> core.Translator:
        translator = core.Translator.for_language(language)
        cls.installTranslator(translator)
        cls.translators[language] = translator
        return translator

    def post_event(
        self,
        obj: QtCore.QObject,
        event: QtCore.QEvent,
        priority: Union[int, constants.EventPriorityStr] = "normal",
    ):
        if isinstance(priority, int):
            prio = priority
        else:
            if priority not in constants.EVENT_PRIORITY:
                raise InvalidParamError(priority, constants.EVENT_PRIORITY)
            prio = constants.EVENT_PRIORITY[priority]
        return self.postEvent(obj, event, prio)

    def main_loop(self) -> int:
        return self.exec_()

    @staticmethod
    def restart():
        os.execl(sys.executable, sys.executable, *sys.argv)



from __future__ import annotations

from prettyqt import constants, gui, widgets
from prettyqt.qt import QtWidgets
from prettyqt.utils import InvalidParamError


QtWidgets.QGraphicsPolygonItem.__bases__ = (widgets.AbstractGraphicsShapeItem,)


class GraphicsPolygonItem(QtWidgets.QGraphicsPolygonItem):
    def serialize_fields(self):
        return dict(polygon=self.get_polygon(), fill_rule=self.get_fill_rule())

    def get_polygon(self) -> gui.PolygonF:
        return gui.PolygonF(self.polygon())

    def set_fill_rule(self, rule: constants.FillRuleStr):
        if rule not in constants.FILL_RULE:
            raise InvalidParamError(rule, constants.FILL_RULE)
        self.setFillRule(constants.FILL_RULE[rule])

    def get_fill_rule(self) -> constants.FillRuleStr:
        return constants.FILL_RULE.inverse[self.fillRule()]



from __future__ import annotations

from typing import List, Literal, Optional

from prettyqt import widgets
from prettyqt.qt import QtWidgets
from prettyqt.utils import InvalidParamError, bidict, types


INPUT_DIALOG_OPTION = bidict(
    no_buttons=QtWidgets.QInputDialog.NoButtons,
    use_listview_for_combobox_items=QtWidgets.QInputDialog.UseListViewForComboBoxItems,
    use_plaintextedit_for_text_input=QtWidgets.QInputDialog.UsePlainTextEditForTextInput,
)

InputDialogOptionStr = Literal[
    "no_buttons", "use_listview_for_combobox_items", "use_plaintextedit_for_text_input"
]

INPUT_MODE = bidict(
    text=QtWidgets.QInputDialog.TextInput,
    int=QtWidgets.QInputDialog.IntInput,
    double=QtWidgets.QInputDialog.DoubleInput,
)

InputModeStr = Literal["text", "int", "double"]

QtWidgets.QInputDialog.__bases__ = (widgets.BaseDialog,)


class InputDialog(QtWidgets.QInputDialog):
    @classmethod
    def get_int(
        cls,
        title: str = "",
        label: str = "",
        icon: types.IconType = None,
    ) -> Optional[int]:
        par = widgets.Dialog()
        par.set_icon(icon)
        v = cls.getInt(par, title, label, value=0)
        return v[0] if v[1] else None

    @classmethod
    def get_float(
        cls,
        title: str = "",
        label: str = "",
        icon: types.IconType = None,
    ) -> Optional[float]:
        par = widgets.Dialog()
        par.set_icon(icon)
        v = cls.getDouble(par, title, label, value=0.0)
        return v[0] if v[1] else None

    @classmethod
    def get_text(
        cls,
        title: str = "",
        label: str = "",
        icon: types.IconType = None,
        default_value: str = "",
        echo_mode: widgets.lineedit.EchoModeStr = "normal",
    ) -> Optional[str]:
        par = widgets.Dialog()
        par.set_icon(icon)
        if echo_mode not in widgets.lineedit.ECHO_MODE:
            raise InvalidParamError(echo_mode, widgets.lineedit.ECHO_MODE)
        v = cls.getText(
            par, title, label, widgets.lineedit.ECHO_MODE[echo_mode], default_value
        )
        return v[0] if v[1] else None

    @classmethod
    def get_item(
        cls,
        items: List[str],
        title: str = "",
        label: str = "",
        icon: types.IconType = None,
        editable: bool = False,
    ) -> Optional[str]:
        par = widgets.Dialog()
        par.set_icon(icon)
        v = cls.getItem(par, title, label, items, editable=editable, current=0)
        return v[0] if v[1] else None

    def set_text_echo_mode(self, mode: widgets.lineedit.EchoModeStr):
        """Set text echo mode.

        Args:
            mode: echo mode to use

        Raises:
            InvalidParamError: invalid echo mode
        """
        if mode not in widgets.lineedit.ECHO_MODE:
            raise InvalidParamError(mode, widgets.lineedit.ECHO_MODE)
        self.setTextEchoMode(widgets.lineedit.ECHO_MODE[mode])

    def get_text_echo_mode(self) -> widgets.lineedit.EchoModeStr:
        """Return text echo mode.

        Returns:
            echo mode
        """
        return widgets.lineedit.ECHO_MODE.inverse[self.textEchoMode()]

    def set_input_mode(self, mode: InputModeStr):
        """Set input mode.

        Args:
            mode: input mode to use

        Raises:
            InvalidParamError: invalid input mode
        """
        if mode not in INPUT_MODE:
            raise InvalidParamError(mode, INPUT_MODE)
        self.setInputMode(INPUT_MODE[mode])

    def get_input_mode(self) -> InputModeStr:
        """Return input mode.

        Returns:
            input mode
        """
        return INPUT_MODE.inverse[self.inputMode()]


if __name__ == "__main__":
    app = widgets.app()
    result = InputDialog.get_int("a", "b", icon="mdi.timer")
    print(result)
    app.main_loop()



from __future__ import annotations

import contextlib
import logging
import os
import pathlib
from typing import Iterator, Literal, Mapping, MutableMapping, Optional, Union

import qstylizer.parser
import qstylizer.style

import prettyqt
from prettyqt import constants, core, gui, widgets
from prettyqt.qt import QtCore, QtWidgets
from prettyqt.utils import InvalidParamError


logger = logging.getLogger(__name__)


SAVE_STATES = dict(
    splitters=QtWidgets.QSplitter,
    mainwindows=QtWidgets.QMainWindow,
    headerviews=QtWidgets.QHeaderView,
)

QtWidgets.QApplication.__bases__ = (gui.GuiApplication,)


class Application(QtWidgets.QApplication):
    def __class_getitem__(cls, name: str) -> QtWidgets.QWidget:
        widget = cls.get_widget(name)
        if widget is None:
            raise ValueError(f"Widget {name!r} does not exist.")
        return widget

    def __iter__(self) -> Iterator[QtWidgets.QWidget]:
        return iter(self.topLevelWidgets())

    def serialize_fields(self):
        return dict(
            auto_sip_enabled=self.autoSipEnabled(),
            cursor_flash_time=self.cursorFlashTime(),
            double_click_interval=self.doubleClickInterval(),
            keyboard_input_interval=self.keyboardInputInterval(),
            start_drag_distance=self.startDragDistance(),
            start_drag_time=self.startDragTime(),
            style_sheet=self.styleSheet(),
            wheel_scroll_lines=self.wheelScrollLines(),
        )

    def store_widget_states(
        self, settings: Optional[MutableMapping] = None, key: str = "states"
    ):
        settings = core.Settings() if settings is None else settings
        result = {}
        for k, v in SAVE_STATES.items():
            result[k] = {
                i.objectName(): i.saveState()
                for i in self.find_children(v)
                if i.objectName()
            }
        settings[key] = result

    def restore_widget_states(
        self, settings: Optional[Mapping] = None, key: str = "states"
    ):
        settings = core.Settings() if settings is None else settings
        for category, v in SAVE_STATES.items():
            items = settings[key].get(category)
            if items is None:
                continue
            for name, state in items.items():
                w = self.find_child(v, name=name)
                if w is not None:
                    w.restoreState(state)

    def about_popup(self, title: str = "About"):
        text = (
            f"{self.applicationName()}\n\n"
            f"{self.organizationName()}\n"
            f"{self.applicationVersion()}\n"
            f"{self.organizationDomain()}"
        )
        popup = widgets.MessageBox(
            widgets.MessageBox.NoIcon, title, text, buttons=widgets.MessageBox.Ok
        )
        popup.set_icon("mdi.information-outline")
        popup.exec_()

    @classmethod
    def get_mainwindow(cls) -> Optional[QtWidgets.QMainWindow]:
        widget_list = cls.instance().topLevelWidgets()
        for widget in widget_list:
            if isinstance(widget, QtWidgets.QMainWindow):
                return widget
        return None

    @classmethod
    def get_widget(cls, name: str) -> Optional[QtWidgets.QWidget]:
        mw = cls.get_mainwindow()
        if mw is None:
            logger.warning("Trying to get widget from nonexistent mainwindow")
            return None
        return mw.findChild(QtWidgets.QWidget, name)  # type: ignore
        # widget_list = cls.instance().allWidgets()
        # for widget in widget_list:
        #     if isinstance(widget, QtWidgets.QWidget) and widget.objectName() == name:
        #         return widget
        # return None

    @contextlib.contextmanager
    def edit_stylesheet(self) -> Iterator[qstylizer.style.StyleSheet]:
        ss = self.get_stylesheet()
        yield ss
        self.set_stylesheet(ss)

    def set_stylesheet(
        self, ss: Union[None, str, qstylizer.style.StyleSheet, os.PathLike]
    ):
        if isinstance(ss, os.PathLike):  # type: ignore
            ss = pathlib.Path(ss).read_text()
        elif ss is None:
            ss = ""
        self.setStyleSheet(str(ss))

    def get_stylesheet(self) -> qstylizer.style.StyleSheet:
        return qstylizer.parser.parse(self.styleSheet())

    def set_theme(self, theme: Literal["default", "dark"]):
        if theme == "default":
            self.set_stylesheet("")
        elif theme == "dark":
            ss = (prettyqt.ROOT_PATH / "themes" / "darktheme.qss").read_text()
            self.set_stylesheet(ss)

    def send_event(self, obj_or_str: Union[str, QtCore.QObject], event: QtCore.QEvent):
        obj = self.get_widget(obj_or_str) if isinstance(obj_or_str, str) else obj_or_str
        if obj is None:
            raise ValueError(obj)
        return self.sendEvent(obj, event)

    def post_event(
        self,
        obj_or_str: Union[str, QtCore.QObject],
        event: QtCore.QEvent,
        priority: Union[int, constants.EventPriorityStr] = "normal",
    ):
        obj = self.get_widget(obj_or_str) if isinstance(obj_or_str, str) else obj_or_str
        if obj is None:
            raise ValueError(obj)
        super().post_event(obj, event, priority)

    @classmethod
    def get_style_icon(cls, icon: str) -> gui.Icon:
        style = cls.style()
        # icon_size = style.pixelMetric(QtWidgets.QStyle.PM_MessageBoxIconSize)
        if icon not in widgets.style.STANDARD_PIXMAP:
            raise InvalidParamError(icon, widgets.style.STANDARD_PIXMAP)
        icon = style.standardIcon(widgets.style.STANDARD_PIXMAP[icon])
        return gui.Icon(icon)

    def set_effect_enabled(self, effect: constants.UiEffectStr, enabled: bool = True):
        """Set the enabled state of a desktop effect.

        Args:
            effect: desktop effect to set
            enabled: new state

        Raises:
            InvalidParamError: invalid desktop effect
        """
        if effect not in constants.UI_EFFECTS:
            raise InvalidParamError(effect, constants.UI_EFFECTS)
        self.setEffectEnabled(constants.UI_EFFECTS[effect])

    def is_effect_enabled(self, effect: constants.UiEffectStr) -> bool:
        """Return desktop effect state.

        Returns:
            desktop effect state
        """
        return self.isEffectEnabled(constants.UI_EFFECTS[effect])

    def set_navigation_mode(self, mode: constants.NavigationModeStr):
        """Set the navigation mode.

        Args:
            mode: navigation mode to use

        Raises:
            InvalidParamError: invalid navigation mode
        """
        if mode not in constants.NAVIGATION_MODES:
            raise InvalidParamError(mode, constants.NAVIGATION_MODES)
        self.setNavigationMode(constants.NAVIGATION_MODES[mode])

    def get_navigation_mode(self) -> constants.NavigationModeStr:
        """Return navigation mode.

        Returns:
            navigation mode
        """
        return constants.NAVIGATION_MODES.inverse[self.navigationMode()]


if __name__ == "__main__":
    app = Application([])
    app.set_theme("dark")
    app.load_language("de")



from __future__ import annotations

from prettyqt import widgets
from prettyqt.qt import QtWidgets


QtWidgets.QStyleOptionTabWidgetFrame.__bases__ = (widgets.StyleOption,)


class StyleOptionTabWidgetFrame(QtWidgets.QStyleOptionTabWidgetFrame):
    pass



from __future__ import annotations

from prettyqt import core
from prettyqt.qt import QtBluetooth
from prettyqt.utils import bidict


md = QtBluetooth.QBluetoothUuid

CHARACTERISTIC_TYPES = bidict(
    aerobic_heart_rate_lower=md.AerobicHeartRateLowerLimit,
    aerobic_heart_rate_upper=md.AerobicHeartRateUpperLimit,
    aerobic_threshold=md.AerobicThreshold,
    age=md.Age,
    anaerobic_heart_rate_lower=md.AnaerobicHeartRateLowerLimit,
    anaerobic_heart_rate_upper=md.AnaerobicHeartRateUpperLimit,
    anaerobic_threshold=md.AnaerobicThreshold,
    alert_category_id=md.AlertCategoryID,
    alert_category_id_bitmask=md.AlertCategoryIDBitMask,
    alert_level=md.AlertLevel,
    alert_notification_control_point=md.AlertNotificationControlPoint,
    alert_status=md.AlertStatus,
    apparent_wind_direction=md.ApparentWindDirection,
    apparent_wind_speed=md.ApparentWindSpeed,
    appearance=md.Appearance,
    barometric_pressure_trend=md.BarometricPressureTrend,
    battery_level=md.BatteryLevel,
    blood_pressure_feature=md.BloodPressureFeature,
    blood_pressure_measurement=md.BloodPressureMeasurement,
    body_composition_feature=md.BodyCompositionFeature,
    body_composition_measurement=md.BodyCompositionMeasurement,
    body_sensor_location=md.BodySensorLocation,
    boot_keyboard_input_report=md.BootKeyboardInputReport,
    boot_keyboard_output_report=md.BootKeyboardOutputReport,
    boot_mouse_input_report=md.BootMouseInputReport,
    csc_feature=md.CSCFeature,
    csc_measurement=md.CSCMeasurement,
    current_time=md.CurrentTime,
    cycling_power_control_point=md.CyclingPowerControlPoint,
    cycling_power_feature=md.CyclingPowerFeature,
    cycling_power_measurement=md.CyclingPowerMeasurement,
    cycling_power_vector=md.CyclingPowerVector,
    database_change_increment=md.DatabaseChangeIncrement,
    date_of_birth=md.DateOfBirth,
    date_of_threshold_assessment=md.DateOfThresholdAssessment,
    date_time=md.DateTime,
    day_date_time=md.DayDateTime,
    day_of_week=md.DayOfWeek,
    descriptor_value_changed=md.DescriptorValueChanged,
    device_name=md.DeviceName,
    dew_point=md.DewPoint,
    dst_offset=md.DSTOffset,
    elevation=md.Elevation,
    email_address=md.EmailAddress,
    exact_time_256=md.ExactTime256,
    fat_burn_hear_rate_lower_limit=md.FatBurnHeartRateLowerLimit,
    fat_burn_hear_rate_upper_limit=md.FatBurnHeartRateUpperLimit,
    firmware_revision_string=md.FirmwareRevisionString,
    first_name=md.FirstName,
    five_zone_heart_rate_limits=md.FiveZoneHeartRateLimits,
    gender=md.Gender,
    glucose_feature=md.GlucoseFeature,
    glucose_measurement=md.GlucoseMeasurement,
    glucose_measurement_context=md.GlucoseMeasurementContext,
    gust_factor=md.GustFactor,
    hardware_revision_string=md.HardwareRevisionString,
    max_recommended_heart_rate=md.MaximumRecommendedHeartRate,
    heart_rate_control_point=md.HeartRateControlPoint,
    heart_rate_max=md.HeartRateMax,
    heart_rate_measurement=md.HeartRateMeasurement,
    heat_index=md.HeatIndex,
    height=md.Height,
    hid_control_point=md.HIDControlPoint,
    hid_information=md.HIDInformation,
    hip_circum_reference=md.HipCircumference,
    humidity=md.Humidity,
    regulatory_certification_list=md.IEEE1107320601RegulatoryCertificationDataList,
    intermediate_cuff_pressure=md.IntermediateCuffPressure,
    intermediate_temperature=md.IntermediateTemperature,
    irradiance=md.Irradiance,
    language=md.Language,
    last_name=md.LastName,
    ln_control_point=md.LNControlPoint,
    ln_feature=md.LNFeature,
    local_time_information=md.LocalTimeInformation,
    location_and_speed=md.LocationAndSpeed,
    magnetic_declination=md.MagneticDeclination,
    magnetic_flux_density_2d=md.MagneticFluxDensity2D,
    magnetic_magnetic_flux_density_3d=md.MagneticFluxDensity3D,
    manufacturer_name_string=md.ManufacturerNameString,
    measurement_interval=md.MeasurementInterval,
    model_number_string=md.ModelNumberString,
    navigation=md.Navigation,
    new_alert=md.NewAlert,
    peripheral_preferred_connection_parameters=md.PeripheralPreferredConnectionParameters,
    peripheral_privacy_flag=md.PeripheralPrivacyFlag,
    pn_pid=md.PnPID,
    pollen_concentration=md.PollenConcentration,
    position_quality=md.PositionQuality,
    pressure=md.Pressure,
    protocol_mode=md.ProtocolMode,
    rainfall=md.Rainfall,
    reconnection_address=md.ReconnectionAddress,
    record_access_control_point=md.RecordAccessControlPoint,
    reference_time_information=md.ReferenceTimeInformation,
    report=md.Report,
    report_map=md.ReportMap,
    resting_heart_rate=md.RestingHeartRate,
    ringer_control_point=md.RingerControlPoint,
    ringer_setting=md.RingerSetting,
    rsc_feature=md.RSCFeature,
    rsc_measurement=md.RSCMeasurement,
    sc_control_point=md.SCControlPoint,
    scan_interval_window=md.ScanIntervalWindow,
    scan_refresh=md.ScanRefresh,
    sensor_location=md.SensorLocation,
    serial_number_string=md.SerialNumberString,
    service_changed=md.ServiceChanged,
    software_revision_string=md.SoftwareRevisionString,
    sport_type_for_aerobic_anaerobic_thresholds=md.SportTypeForAerobicAnaerobicThresholds,
    supported_new_alert_category=md.SupportedNewAlertCategory,
    supported_unread_alert_category=md.SupportedUnreadAlertCategory,
    system_id=md.SystemID,
    temperature=md.Temperature,
    temperature_measurement=md.TemperatureMeasurement,
    temperature_type=md.TemperatureType,
    three_zone_heart_rate_limits=md.ThreeZoneHeartRateLimits,
    time_accuracy=md.TimeAccuracy,
    time_source=md.TimeSource,
    time_update_control_point=md.TimeUpdateControlPoint,
    time_update_state=md.TimeUpdateState,
    time_with_dst=md.TimeWithDST,
    time_zone=md.TimeZone,
    true_wind_direction=md.TrueWindDirection,
    true_wind_speed=md.TrueWindSpeed,
    two_zone_hear_rate_limits=md.TwoZoneHeartRateLimits,
    tx_power_level=md.TxPowerLevel,
    unread_alert_status=md.UnreadAlertStatus,
    user_control_point=md.UserControlPoint,
    user_index=md.UserIndex,
    uv_index=md.UVIndex,
    vo_2_max=md.VO2Max,
    waist_circumference=md.WaistCircumference,
    weight=md.Weight,
    weight_measurement=md.WeightMeasurement,
    weight_scale_feature=md.WeightScaleFeature,
    wind_chill=md.WindChill,
)

DESCRIPTOR_TYPE = bidict(
    characteristic_extended_properties=md.CharacteristicExtendedProperties,
    characteristic_user_description=md.CharacteristicUserDescription,
    client_characteristic_configuration=md.ClientCharacteristicConfiguration,
    server_characteristic_configuration=md.ServerCharacteristicConfiguration,
    characteristic_presentation_format=md.CharacteristicPresentationFormat,
    characteristic_aggregate_format=md.CharacteristicAggregateFormat,
    valid_range=md.ValidRange,
    external_report_reference=md.ExternalReportReference,
    report_reference=md.ReportReference,
    environmental_sensing_configuration=md.EnvironmentalSensingConfiguration,
    environmental_sensing_measurement=md.EnvironmentalSensingMeasurement,
    environmental_sensing_trigger_setting=md.EnvironmentalSensingTriggerSetting,
    unknown_descriptor_type=md.UnknownDescriptorType,
)

PROTOCOL_UUID = bidict(
    sdp=md.Sdp,
    udp=md.Udp,
    rfcomm=md.Rfcomm,
    tcp=md.Tcp,
    tcs_bin=md.TcsBin,
    tcs_at=md.TcsAt,
    att=md.Att,
    obex=md.Obex,
    ip=md.Ip,
    ftp=md.Ftp,
    http=md.Http,
    wsp=md.Wsp,
    bnep=md.Bnep,
    upnp=md.Upnp,
    hidp=md.Hidp,
    hardcopy_control_channel=md.HardcopyControlChannel,
    hardcopy_data_channel=md.HardcopyDataChannel,
    hardcopy_notification=md.HardcopyNotification,
    avctp=md.Avctp,
    avdtp=md.Avdtp,
    cmtp=md.Cmtp,
    udi_c_plain=md.UdiCPlain,
    mcap_control_channel=md.McapControlChannel,
    mcap_data_channel=md.McapDataChannel,
    l2_cap=md.L2cap,
)

SERVICE_CLASS_UUID = bidict(
    service_discovery_server=md.ServiceDiscoveryServer,
    browser_group_descriptor=md.BrowseGroupDescriptor,
    public_browse_group=md.PublicBrowseGroup,
    serial_port=md.SerialPort,
    lan_access_using_ppp=md.LANAccessUsingPPP,
    dial_up_networking=md.DialupNetworking,
    ir_mc_sync=md.IrMCSync,
    obex_object_push=md.ObexObjectPush,
    obex_file_transfer=md.OBEXFileTransfer,
    ir_mc_sync_command=md.IrMCSyncCommand,
    headset=md.Headset,
    audio_source=md.AudioSource,
    audio_sink=md.AudioSink,
    av_remove_control_target=md.AV_RemoteControlTarget,
    advanced_audio_distribution=md.AdvancedAudioDistribution,
    av_remote_control=md.AV_RemoteControl,
    av_remote_control_controller=md.AV_RemoteControlController,
    headset_ag=md.HeadsetAG,
    panu=md.PANU,
    nap=md.NAP,
    gn=md.GN,
    direct_printing=md.DirectPrinting,
    reference_printing=md.ReferencePrinting,
    basic_image=md.BasicImage,
    imaging_responder=md.ImagingResponder,
    imaging_automatic_archive=md.ImagingAutomaticArchive,
    imaging_reference_objects=md.ImagingReferenceObjects,
    hands_free=md.Handsfree,
    hands_free_audio_gateway=md.HandsfreeAudioGateway,
    direct_printing_reference_objects_service=md.DirectPrintingReferenceObjectsService,
    reflected_ui=md.ReflectedUI,
    basic_printing=md.BasicPrinting,
    printing_status=md.PrintingStatus,
    human_interface_device_service=md.HumanInterfaceDeviceService,
    hardcopy_cable_replacement=md.HardcopyCableReplacement,
    hcr_print=md.HCRPrint,
    hcr_scan=md.HCRScan,
    sim_access=md.SIMAccess,
    phonebook_access_pce=md.PhonebookAccessPCE,
    phonebook_access_pse=md.PhonebookAccessPSE,
    phonebook_access=md.PhonebookAccess,
    headset_hs=md.HeadsetHS,
    message_access_server=md.MessageAccessServer,
    message_notification_server=md.MessageNotificationServer,
    message_access_profile=md.MessageAccessProfile,
    gnss=md.GNSS,
    gnss_server=md.GNSSServer,
    display_3d=md.Display3D,
    glasses_3d=md.Glasses3D,
    synchronization_3d=md.Synchronization3D,
    mps_profile=md.MPSProfile,
    mps_service=md.MPSService,
    pnp_information=md.PnPInformation,
    generic_networking=md.GenericNetworking,
    generic_file_transfer=md.GenericFileTransfer,
    generic_audio=md.GenericAudio,
    generic_telephony=md.GenericTelephony,
    video_source=md.VideoSource,
    video_sink=md.VideoSink,
    video_distribution=md.VideoDistribution,
    hdp=md.HDP,
    hdp_source=md.HDPSource,
    hdp_sink=md.HDPSink,
    generic_access=md.GenericAccess,
    generic_attribute=md.GenericAttribute,
    immediate_alert=md.ImmediateAlert,
    link_loss=md.LinkLoss,
    tx_power=md.TxPower,
    current_time_service=md.CurrentTimeService,
    reference_time_update_service=md.ReferenceTimeUpdateService,
    next_dst_change_service=md.NextDSTChangeService,
    glucose=md.Glucose,
    health_thermometer=md.HealthThermometer,
    device_information=md.DeviceInformation,
    heart_rate=md.HeartRate,
    phone_alert_status_service=md.PhoneAlertStatusService,
    battery_service=md.BatteryService,
    blood_pressure=md.BloodPressure,
    alert_notification_service=md.AlertNotificationService,
    human_interface_device=md.HumanInterfaceDevice,
    scan_parameters=md.ScanParameters,
    running_speed_and_cadence=md.RunningSpeedAndCadence,
    cycling_speed_and_cadence=md.CyclingSpeedAndCadence,
    cycling_power=md.CyclingPower,
    location_and_navigation=md.LocationAndNavigation,
    environmental_sensing=md.EnvironmentalSensing,
    body_composition=md.BodyComposition,
    user_data=md.UserData,
    weight_scale=md.WeightScale,
    bond_management=md.BondManagement,
    continuous_glucose_monitoring=md.ContinuousGlucoseMonitoring,
)

QtBluetooth.QBluetoothUuid.__bases__ = (core.Uuid,)


class BluetoothUuid(QtBluetooth.QBluetoothUuid):
    pass


if __name__ == "__main__":
    address = BluetoothUuid()



from __future__ import annotations

from typing import List, Literal, Optional

from prettyqt import core, widgets
from prettyqt.qt import QtCore, QtWidgets


SideStr = Literal["left", "top", "right", "bottom"]


class SideGrip(widgets.Widget):
    def __init__(
        self,
        parent: Optional[QtWidgets.QWidget],
        edge: SideStr,
    ):
        super().__init__(parent=parent)
        if edge == "left":
            self.set_cursor("size_horizonal")
            self.resize_fn = self.resize_left
        elif edge == "top":
            self.set_cursor("size_vertical")
            self.resize_fn = self.resize_top
        elif edge == "right":
            self.set_cursor("size_horizonal")
            self.resize_fn = self.resize_right
        else:
            self.set_cursor("size_vertical")
            self.resize_fn = self.resize_bottom
        self.mouse_pos = None

    def resize_left(self, delta: QtCore.QPoint):
        window = self.window()
        width = max(window.minimumWidth(), window.width() - delta.x())
        geo = window.geometry()
        geo.setLeft(geo.right() - width)
        window.setGeometry(geo)

    def resize_top(self, delta: QtCore.QPoint):
        window = self.window()
        height = max(window.minimumHeight(), window.height() - delta.y())
        geo = window.geometry()
        geo.setTop(geo.bottom() - height)
        window.setGeometry(geo)

    def resize_right(self, delta: QtCore.QPoint):
        window = self.window()
        width = max(window.minimumWidth(), window.width() + delta.x())
        window.resize(width, window.height())

    def resize_bottom(self, delta: QtCore.QPoint):
        window = self.window()
        height = max(window.minimumHeight(), window.height() + delta.y())
        window.resize(window.width(), height)

    def mousePressEvent(self, event):
        if event.button() == QtCore.Qt.LeftButton:
            self.mouse_pos = event.pos()

    def mouseMoveEvent(self, event):
        if self.mouse_pos is not None:
            delta = event.pos() - self.mouse_pos
            self.resize_fn(delta)

    def mouseReleaseEvent(self, event):
        self.mouse_pos = None


class FramelessWindow(widgets.MainWindow):
    _grip_size = 4

    def __init__(self, parent: Optional[QtWidgets.QWidget] = None):
        super().__init__(parent=parent)
        self.set_flags(frameless=True)
        sides: List[SideStr] = ["left", "top", "right", "bottom"]
        self.side_grips = [SideGrip(self, side) for side in sides]
        # corner grips should be "on top" of everything, otherwise the side grips
        # will take precedence on mouse events, so we are adding them *after*;
        # alternatively, widget.raise_() can be used
        self.corner_grips = [widgets.SizeGrip(self) for i in range(4)]
        self.set_margin(self.grip_size)

    @property
    def grip_size(self) -> int:
        return self._grip_size

    def set_grip_size(self, size: int):
        if size == self._grip_size:
            return
        self._grip_size = max(2, size)
        self.set_margin(self.grip_size)
        self.update_grips()

    def update_grips(self):
        outer = self.rect()
        # an "inner" rect used for reference to set the geometries of size grips
        inner = outer.adjusted(
            self.grip_size, self.grip_size, -self.grip_size, -self.grip_size
        )

        # top left
        rect = core.Rect(outer.topLeft(), inner.topLeft())
        self.corner_grips[0].setGeometry(rect)
        # top right
        rect = core.Rect(outer.topRight(), inner.topRight()).normalized()
        self.corner_grips[1].setGeometry(rect)
        # bottom right
        rect = core.Rect(inner.bottomRight(), outer.bottomRight())
        self.corner_grips[2].setGeometry(rect)
        # bottom left
        rect = core.Rect(outer.bottomLeft(), inner.bottomLeft()).normalized()
        self.corner_grips[3].setGeometry(rect)

        # left edge
        self.side_grips[0].setGeometry(0, inner.top(), self.grip_size, inner.height())
        # top edge
        self.side_grips[1].setGeometry(inner.left(), 0, inner.width(), self.grip_size)
        # right edge
        self.side_grips[2].setGeometry(
            inner.right(),
            inner.top(),
            self.grip_size,
            inner.height(),
        )
        # bottom edge
        self.side_grips[3].setGeometry(
            self.grip_size,
            inner.bottom(),
            inner.width(),
            self.grip_size,
        )

    def resizeEvent(self, event):
        super().resizeEvent(event)
        self.update_grips()


if __name__ == "__main__":
    app = widgets.app()
    m = FramelessWindow()
    button = widgets.PushButton("test")
    m.set_widget(button)
    m.show()
    m.resize(240, 160)
    app.exec_()



from __future__ import annotations

from typing import Literal

from prettyqt import core
from prettyqt.qt import QtCore
from prettyqt.utils import InvalidParamError, bidict


CHILD_MODE = bidict(
    exclusive=QtCore.QState.ExclusiveStates, parallel=QtCore.QState.ParallelStates
)

ChildModeStr = Literal["exclusive", "parallel"]

RESTORE_POLICY = bidict(
    dont_restore=QtCore.QState.DontRestoreProperties,
    restore=QtCore.QState.RestoreProperties,
)

RestorePolicyStr = Literal["dont_restore", "restore"]


QtCore.QState.__bases__ = (core.AbstractState,)


class State(QtCore.QState):
    def set_child_mode(self, mode: ChildModeStr):
        """Set child mode to use.

        Args:
            mode: child mode to use

        Raises:
            InvalidParamError: child mode does not exist
        """
        if mode not in CHILD_MODE:
            raise InvalidParamError(mode, CHILD_MODE)
        self.setChildMode(CHILD_MODE[mode])

    def get_child_mode(self) -> ChildModeStr:
        """Return current child mode.

        Returns:
            child mode
        """
        return CHILD_MODE.inverse[self.childMode()]


if __name__ == "__main__":
    reg = State()



from __future__ import annotations

from typing import Optional

from prettyqt import widgets
from prettyqt.qt import QtCore, QtWidgets


QtWidgets.QWidgetAction.__bases__ = (widgets.Action,)


class WidgetAction(QtWidgets.QWidgetAction):
    def __init__(self, parent: Optional[QtCore.QObject] = None):
        super().__init__(parent)  # type: ignore
        self._menu = None  # bc of PySide inheritance also defined here

    #     self.set_text(text)
    #     self.set_icon(icon)
    #     self.set_shortcut(shortcut)
    #     self.set_tooltip(tooltip)


if __name__ == "__main__":
    w = WidgetAction()



from __future__ import annotations

from prettyqt.qt import QtCore
from prettyqt.utils import bidict


TYPE = bidict(
    none=QtCore.QEvent.None_,
    action_added=QtCore.QEvent.ActionAdded,
    action_changed=QtCore.QEvent.ActionChanged,
    action_removed=QtCore.QEvent.ActionRemoved,
    activation_change=QtCore.QEvent.ActivationChange,
    application_activate=QtCore.QEvent.ApplicationActivate,
    application_deactivate=QtCore.QEvent.ApplicationDeactivate,
    application_font_change=QtCore.QEvent.ApplicationFontChange,
    application_layout_direction_change=QtCore.QEvent.ApplicationLayoutDirectionChange,
    application_palette_change=QtCore.QEvent.ApplicationPaletteChange,
    application_state_change=QtCore.QEvent.ApplicationStateChange,
    application_window_icon_change=QtCore.QEvent.ApplicationWindowIconChange,
    child_added=QtCore.QEvent.ChildAdded,
    child_polished=QtCore.QEvent.ChildPolished,
    child_removed=QtCore.QEvent.ChildRemoved,
    clipboard=QtCore.QEvent.Clipboard,
    close=QtCore.QEvent.Close,
    close_software_input_panel=QtCore.QEvent.CloseSoftwareInputPanel,
    contents_rect_change=QtCore.QEvent.ContentsRectChange,
    context_menu=QtCore.QEvent.ContextMenu,
    cursor_change=QtCore.QEvent.CursorChange,
    deferred_delete=QtCore.QEvent.DeferredDelete,
    drag_enter=QtCore.QEvent.DragEnter,
    drag_leave=QtCore.QEvent.DragLeave,
    drag_move=QtCore.QEvent.DragMove,
    drop=QtCore.QEvent.Drop,
    dynamic_property_change=QtCore.QEvent.DynamicPropertyChange,
    enabled_change=QtCore.QEvent.EnabledChange,
    enter=QtCore.QEvent.Enter,
    # enter_edit_focus=QtCore.QEvent.EnterEditFocus,
    enter_whats_this_mode=QtCore.QEvent.EnterWhatsThisMode,
    expose=QtCore.QEvent.Expose,
    file_open=QtCore.QEvent.FileOpen,
    focus_in=QtCore.QEvent.FocusIn,
    focus_out=QtCore.QEvent.FocusOut,
    focus_about_to_change=QtCore.QEvent.FocusAboutToChange,
    font_change=QtCore.QEvent.FontChange,
    gesture=QtCore.QEvent.Gesture,
    gesture_override=QtCore.QEvent.GestureOverride,
    grab_keyboard=QtCore.QEvent.GrabKeyboard,
    grab_mouse=QtCore.QEvent.GrabMouse,
    graphics_scene_context_menu=QtCore.QEvent.GraphicsSceneContextMenu,
    graphics_scene_drag_enter=QtCore.QEvent.GraphicsSceneDragEnter,
    graphics_scene_drag_leave=QtCore.QEvent.GraphicsSceneDragLeave,
    graphics_scene_drag_move=QtCore.QEvent.GraphicsSceneDragMove,
    graphics_scene_drop=QtCore.QEvent.GraphicsSceneDrop,
    graphics_scene_help=QtCore.QEvent.GraphicsSceneHelp,
    graphics_scene_hover_enter=QtCore.QEvent.GraphicsSceneHoverEnter,
    graphics_scene_hover_leave=QtCore.QEvent.GraphicsSceneHoverLeave,
    graphics_scene_hover_move=QtCore.QEvent.GraphicsSceneHoverMove,
    graphics_scene_mouse_double_click=QtCore.QEvent.GraphicsSceneMouseDoubleClick,
    graphics_scene_mouse_move=QtCore.QEvent.GraphicsSceneMouseMove,
    graphics_scene_mouse_press=QtCore.QEvent.GraphicsSceneMousePress,
    graphics_scene_mouse_release=QtCore.QEvent.GraphicsSceneMouseRelease,
    graphics_scene_move=QtCore.QEvent.GraphicsSceneMove,
    graphics_scene_resize=QtCore.QEvent.GraphicsSceneResize,
    graphics_scene_wheel=QtCore.QEvent.GraphicsSceneWheel,
    hide=QtCore.QEvent.Hide,
    hide_to_parent=QtCore.QEvent.HideToParent,
    hover_enter=QtCore.QEvent.HoverEnter,
    hover_leave=QtCore.QEvent.HoverLeave,
    hover_move=QtCore.QEvent.HoverMove,
    icon_drag=QtCore.QEvent.IconDrag,
    icon_text_change=QtCore.QEvent.IconTextChange,
    input_method=QtCore.QEvent.InputMethod,
    input_method_query=QtCore.QEvent.InputMethodQuery,
    keyboard_layout_change=QtCore.QEvent.KeyboardLayoutChange,
    key_press=QtCore.QEvent.KeyPress,
    key_release=QtCore.QEvent.KeyRelease,
    language_change=QtCore.QEvent.LanguageChange,
    layout_direction_change=QtCore.QEvent.LayoutDirectionChange,
    layout_request=QtCore.QEvent.LayoutRequest,
    leave=QtCore.QEvent.Leave,
    # leave_edit_focus=QtCore.QEvent.LeaveEditFocus,
    leave_whats_this_mode=QtCore.QEvent.LeaveWhatsThisMode,
    locale_change=QtCore.QEvent.LocaleChange,
    non_client_area_mouse_button_dbl_click=QtCore.QEvent.NonClientAreaMouseButtonDblClick,
    non_client_area_mouse_button_press=QtCore.QEvent.NonClientAreaMouseButtonPress,
    non_client_area_mouse_button_release=QtCore.QEvent.NonClientAreaMouseButtonRelease,
    non_client_area_mouse_move=QtCore.QEvent.NonClientAreaMouseMove,
    mac_size_change=QtCore.QEvent.MacSizeChange,
    meta_call=QtCore.QEvent.MetaCall,
    # invoke_method=QtCore.QEvent.invokeMethod,
    modified_change=QtCore.QEvent.ModifiedChange,
    mouse_button_dbl_click=QtCore.QEvent.MouseButtonDblClick,
    mouse_button_press=QtCore.QEvent.MouseButtonPress,
    mouse_button_release=QtCore.QEvent.MouseButtonRelease,
    mouse_move=QtCore.QEvent.MouseMove,
    mouse_tracking_change=QtCore.QEvent.MouseTrackingChange,
    move=QtCore.QEvent.Move,
    # native_gesture=QtCore.QEvent.NativeGesture,
    orientation_change=QtCore.QEvent.OrientationChange,
    paint=QtCore.QEvent.Paint,
    palette_change=QtCore.QEvent.PaletteChange,
    parent_about_to_change=QtCore.QEvent.ParentAboutToChange,
    parent_change=QtCore.QEvent.ParentChange,
    platform_panel=QtCore.QEvent.PlatformPanel,
    platform_surface=QtCore.QEvent.PlatformSurface,
    polish=QtCore.QEvent.Polish,
    polish_request=QtCore.QEvent.PolishRequest,
    query_whats_this=QtCore.QEvent.QueryWhatsThis,
    read_only_change=QtCore.QEvent.ReadOnlyChange,
    request_software_input_panel=QtCore.QEvent.RequestSoftwareInputPanel,
    resize=QtCore.QEvent.Resize,
    scroll_prepare=QtCore.QEvent.ScrollPrepare,
    scroll=QtCore.QEvent.Scroll,
    shortcut=QtCore.QEvent.Shortcut,
    shortcut_override=QtCore.QEvent.ShortcutOverride,
    show=QtCore.QEvent.Show,
    show_to_parent=QtCore.QEvent.ShowToParent,
    sock_act=QtCore.QEvent.SockAct,
    state_machine_signal=QtCore.QEvent.StateMachineSignal,
    # signal_event=QtCore.QEvent.SignalEvent,
    state_machine_wrapped=QtCore.QEvent.StateMachineWrapped,
    # wrapped_event=QtCore.QEvent.WrappedEvent,
    status_tip=QtCore.QEvent.StatusTip,
    style_change=QtCore.QEvent.StyleChange,
    tablet_move=QtCore.QEvent.TabletMove,
    tablet_press=QtCore.QEvent.TabletPress,
    tablet_release=QtCore.QEvent.TabletRelease,
    tablet_enter_proximity=QtCore.QEvent.TabletEnterProximity,
    tablet_leave_proximity=QtCore.QEvent.TabletLeaveProximity,
    tablet_tracking_change=QtCore.QEvent.TabletTrackingChange,
    thread_change=QtCore.QEvent.ThreadChange,
    # move_to_thread=QtCore.QEvent.moveToThread,
    timer=QtCore.QEvent.Timer,
    tool_bar_change=QtCore.QEvent.ToolBarChange,
    tool_tip=QtCore.QEvent.ToolTip,
    tool_tip_change=QtCore.QEvent.ToolTipChange,
    touch_begin=QtCore.QEvent.TouchBegin,
    touch_cancel=QtCore.QEvent.TouchCancel,
    touch_end=QtCore.QEvent.TouchEnd,
    touch_update=QtCore.QEvent.TouchUpdate,
    ungrab_keyboard=QtCore.QEvent.UngrabKeyboard,
    ungrab_mouse=QtCore.QEvent.UngrabMouse,
    update_later=QtCore.QEvent.UpdateLater,
    update_request=QtCore.QEvent.UpdateRequest,
    whats_this=QtCore.QEvent.WhatsThis,
    whats_this_clicked=QtCore.QEvent.WhatsThisClicked,
    wheel=QtCore.QEvent.Wheel,
    win_event_act=QtCore.QEvent.WinEventAct,
    window_activate=QtCore.QEvent.WindowActivate,
    window_blocked=QtCore.QEvent.WindowBlocked,
    window_deactivate=QtCore.QEvent.WindowDeactivate,
    window_icon_change=QtCore.QEvent.WindowIconChange,
    window_state_change=QtCore.QEvent.WindowStateChange,
    window_title_change=QtCore.QEvent.WindowTitleChange,
    window_unblocked=QtCore.QEvent.WindowUnblocked,
    win_id_change=QtCore.QEvent.WinIdChange,
    z_order_change=QtCore.QEvent.ZOrderChange,
)


class Event(QtCore.QEvent):
    def __repr__(self):
        return f"{type(self).__name__}({self.type()})"



from __future__ import annotations

from typing import Literal, Optional

from prettyqt import constants, gui, widgets
from prettyqt.qt import QtWidgets
from prettyqt.utils import InvalidParamError, bidict


SHAPE_MODE = bidict(
    mask=QtWidgets.QGraphicsPixmapItem.MaskShape,
    bounding_rect=QtWidgets.QGraphicsPixmapItem.BoundingRectShape,
    heuristic_mask=QtWidgets.QGraphicsPixmapItem.HeuristicMaskShape,
)

ShapeModeStr = Literal["mask", "bounding_rect", "heuristic_mask"]

QtWidgets.QGraphicsPixmapItem.__bases__ = (widgets.GraphicsItem,)


class GraphicsPixmapItem(QtWidgets.QGraphicsPixmapItem):
    def serialize_fields(self):
        return dict(
            offset=self.offset(),
            pixmap=self.get_pixmap(),
            shape_mode=self.get_shape_mode(),
            transformation_mode=self.get_transformation_mode(),
        )

    def set_transformation_mode(self, mode: constants.TransformationModeStr):
        """Set transformation mode.

        Args:
            mode: transformation mode to use

        Raises:
            InvalidParamError: mode does not exist
        """
        if mode not in constants.TRANSFORMATION_MODE:
            raise InvalidParamError(mode, constants.TRANSFORMATION_MODE)
        self.setTransformationMode(constants.TRANSFORMATION_MODE[mode])

    def get_transformation_mode(self) -> constants.TransformationModeStr:
        """Return current transformation mode.

        Returns:
            transformation mode
        """
        return constants.TRANSFORMATION_MODE.inverse[self.transformationMode()]

    def set_shape_mode(self, mode: ShapeModeStr):
        """Set shape mode.

        Args:
            mode: shape mode to use

        Raises:
            InvalidParamError: mode does not exist
        """
        if mode not in SHAPE_MODE:
            raise InvalidParamError(mode, SHAPE_MODE)
        self.setShapeMode(SHAPE_MODE[mode])

    def get_shape_mode(self) -> ShapeModeStr:
        """Return current shape mode.

        Returns:
            shape mode
        """
        return SHAPE_MODE.inverse[self.shapeMode()]

    def get_pixmap(self) -> Optional[gui.Pixmap]:
        pix = self.pixmap()
        if pix.isNull():
            return None
        return gui.Pixmap(pix)



from __future__ import annotations

from typing import Literal, Union

from prettyqt import gui
from prettyqt.qt import QtGui
from prettyqt.utils import InvalidParamError, bidict, colors, types


FONT_PROPERTY_INHERITANCE_BEHAVIOUR = bidict(
    none=QtGui.QTextCharFormat.FontPropertiesSpecifiedOnly,
    single=QtGui.QTextCharFormat.FontPropertiesAll,
)

UNDERLINE_STYLE = bidict(
    none=QtGui.QTextCharFormat.NoUnderline,
    single=QtGui.QTextCharFormat.SingleUnderline,
    dash=QtGui.QTextCharFormat.DashUnderline,
    dot=QtGui.QTextCharFormat.DotLine,
    dashdot=QtGui.QTextCharFormat.DashDotLine,
    dashdotline=QtGui.QTextCharFormat.DashDotDotLine,
    wave=QtGui.QTextCharFormat.WaveUnderline,
    spellcheck=QtGui.QTextCharFormat.SpellCheckUnderline,
)

UnderlineStyleStr = Literal[
    "none",
    "single",
    "dash",
    "dot",
    "dashdot",
    "dashdotline",
    "wave",
    "spellcheck",
]

VERTICAL_ALIGNMENT = bidict(
    normal=QtGui.QTextCharFormat.AlignNormal,
    super_script=QtGui.QTextCharFormat.AlignSuperScript,
    sub_script=QtGui.QTextCharFormat.AlignSubScript,
    middle=QtGui.QTextCharFormat.AlignMiddle,
    bottom=QtGui.QTextCharFormat.AlignBottom,
    top=QtGui.QTextCharFormat.AlignTop,
    baseline=QtGui.QTextCharFormat.AlignBaseline,
)

VerticalAlignmentStr = Literal[
    "normal",
    "super_script",
    "sub_script",
    "middle",
    "bottom",
    "top",
    "baseline",
]

QtGui.QTextCharFormat.__bases__ = (gui.TextFormat,)


class TextCharFormat(QtGui.QTextCharFormat):
    def __init__(
        self,
        text_color: Union[types.ColorType, QtGui.QBrush] = None,
        bold: bool = False,
        italic: bool = False,
    ):
        super().__init__()
        if text_color is not None:
            self.set_foreground_color(text_color)
        if bold:
            self.set_font_weight("bold")
        self.setFontItalic(italic)

    def set_foreground_color(self, color: Union[types.ColorType, QtGui.QBrush]):
        if not isinstance(color, QtGui.QBrush):
            color = colors.get_color(color)
        self.setForeground(color)

    def set_background_color(self, color: Union[types.ColorType, QtGui.QBrush]):
        if not isinstance(color, QtGui.QBrush):
            color = colors.get_color(color)
        self.setBackground(color)

    def set_font_weight(self, weight: gui.font.WeightStr):
        """Set the font weight.

        Args:
            weight: font weight

        Raises:
            InvalidParamError: invalid font weight
        """
        if weight not in gui.font.WEIGHT:
            raise InvalidParamError(weight, gui.font.WEIGHT)
        self.setFontWeight(gui.font.WEIGHT[weight])

    def get_font_weight(self) -> gui.font.WeightStr:
        """Get current font weight.

        Returns:
            current font weight
        """
        return gui.font.WEIGHT.inverse[self.fontWeight()]

    def set_underline_style(self, style: UnderlineStyleStr):
        """Set the underline style.

        Args:
            style: underline style

        Raises:
            InvalidParamError: invalid underline style
        """
        if style not in UNDERLINE_STYLE:
            raise InvalidParamError(style, UNDERLINE_STYLE)
        self.setUnderlineStyle(UNDERLINE_STYLE[style])

    def get_underline_style(self) -> UnderlineStyleStr:
        """Get current underline style.

        Returns:
            current underline style
        """
        return UNDERLINE_STYLE.inverse[self.underlineStyle()]

    def set_vertical_alignment(self, alignment: VerticalAlignmentStr):
        """Set the vertical alignment.

        Args:
            alignment: vertical alignment

        Raises:
            InvalidParamError: invalid vertical alignment
        """
        if alignment not in VERTICAL_ALIGNMENT:
            raise InvalidParamError(alignment, VERTICAL_ALIGNMENT)
        self.setVerticalAlignment(VERTICAL_ALIGNMENT[alignment])

    def get_vertical_alignment(self) -> VerticalAlignmentStr:
        """Get current vertical alignment.

        Returns:
            current vertical alignment
        """
        return VERTICAL_ALIGNMENT.inverse[self.verticalAlignment()]

    def set_font_style_hint(self, hint: gui.font.StyleHintStr):
        """Set the font style hint.

        Args:
            hint: font style hint

        Raises:
            InvalidParamError: invalid font style hint
        """
        if hint not in gui.font.STYLE_HINTS:
            raise InvalidParamError(hint, gui.font.STYLE_HINTS)
        self.setFontStyleHint(gui.font.STYLE_HINTS[hint])

    def get_font(self) -> gui.Font:
        return gui.Font(self.font())



from __future__ import annotations

from prettyqt import qml
from prettyqt.qt import QtQuick


QtQuick.QQuickImageProvider.__bases__ = (qml.QmlImageProviderBase,)


class QuickImageProvider(QtQuick.QQuickImageProvider):
    pass



from __future__ import annotations

from typing import Optional

from prettyqt import widgets
from prettyqt.qt import QtWidgets


QtWidgets.QProgressDialog.__bases__ = (widgets.BaseDialog,)


class ProgressDialog(QtWidgets.QProgressDialog):
    """Progress dialog.

    Wrapper for QtWidgets.QProgressDialog
    """

    def __init__(self, parent: Optional[QtWidgets.QWidget] = None):
        super().__init__(parent=parent)

        progress_bar = widgets.ProgressBar()
        progress_bar.setRange(0, 0)
        progress_bar.setTextVisible(False)
        self.setBar(progress_bar)

        self.set_icon("mdi.timer-sand-empty")
        self.set_modality("application")
        self.set_flags(
            minimize=False, maximize=False, close=False, stay_on_top=True, window=True
        )
        self.setCancelButton(None)  # type: ignore
        self.cancel()

    def show_message(self, message: str):
        self.setLabelText(message)
        self.show()


if __name__ == "__main__":
    app = widgets.app()
    widget = ProgressDialog()
    widget.show_message("test")
    widget.exec_()



from __future__ import annotations

from typing import Literal, Union

from prettyqt.qt import QtMultimedia
from prettyqt.utils import bidict


HANDLE_TYPE = bidict(
    none=QtMultimedia.QAbstractVideoBuffer.NoHandle,
    gl_texture=QtMultimedia.QAbstractVideoBuffer.GLTextureHandle,
    xv_shm_image=QtMultimedia.QAbstractVideoBuffer.XvShmImageHandle,
    core_image=QtMultimedia.QAbstractVideoBuffer.CoreImageHandle,
    pixmap=QtMultimedia.QAbstractVideoBuffer.QPixmapHandle,
    egl_image=QtMultimedia.QAbstractVideoBuffer.EGLImageHandle,
    user=QtMultimedia.QAbstractVideoBuffer.UserHandle,
)

HandleTypeStr = Literal[
    "none", "gl_texture", "xv_shm_image", "core_image", "pixmap", "egl_image", "user"
]

MAP_MODE = bidict(
    not_mapped=QtMultimedia.QAbstractVideoBuffer.NotMapped,
    read_only=QtMultimedia.QAbstractVideoBuffer.ReadOnly,
    write_only=QtMultimedia.QAbstractVideoBuffer.WriteOnly,
    read_write=QtMultimedia.QAbstractVideoBuffer.ReadWrite,
)

MapModeStr = Literal["not_mapped", "read_only", "write_only", "read_write"]


class AbstractVideoBuffer(QtMultimedia.QAbstractVideoBuffer):
    def __init__(
        self, handle_type: Union[QtMultimedia.QAbstractVideoBuffer.HandleType, str]
    ):
        if isinstance(handle_type, QtMultimedia.QAbstractVideoBuffer.HandleType):
            typ = handle_type
        else:
            typ = HANDLE_TYPE[handle_type]
        super().__init__(typ)

    def __repr__(self):
        return f"{type(self).__name__}({self.get_handle_type()!r})"

    def get_handle_type(self) -> HandleTypeStr:
        """Return current handle type.

        Returns:
            handle type
        """
        return HANDLE_TYPE.inverse[self.handleType()]

    def get_map_mode(self) -> MapModeStr:
        """Return current map mode.

        Returns:
            map mode
        """
        return MAP_MODE.inverse[self.mapMode()]

    def map_planes(
        self,
        mode: Union[int, str],
        num_bytes: int,
        bytes_per_line: int = 4,
        data: int = 4,
    ):
        if mode in MAP_MODE:
            mode = MAP_MODE[mode]
        self.mapPlanes(mode, num_bytes, bytes_per_line, data)

    def map(
        self,
        mode: Union[int, str],
        num_bytes: int,
        bytes_per_line: int = 4,
    ):
        if mode in MAP_MODE:
            mode = MAP_MODE[mode]
        super().map(mode, num_bytes, bytes_per_line)



from __future__ import annotations

from typing import Literal

from prettyqt.qt import QtWebEngineWidgets

# from prettyqt import core
from prettyqt.utils import InvalidParamError, bidict


Item = QtWebEngineWidgets.QWebEngineDownloadItem

DOWNLOAD_INTERRUPT_REASONS = bidict(
    none=Item.NoReason,
    file_failed=Item.FileFailed,
    file_access_denied=Item.FileAccessDenied,
    file_no_space=Item.FileNoSpace,
    filename_too_long=Item.FileNameTooLong,
    file_too_large=Item.FileTooLarge,
    file_virus_infected=Item.FileVirusInfected,
    file_transient_error=Item.FileTransientError,
    file_blocked=Item.FileBlocked,
    file_security_check_failed=Item.FileSecurityCheckFailed,
    file_too_short=Item.FileTooShort,
    file_hash_mismatch=Item.FileHashMismatch,
    network_fialed=Item.NetworkFailed,
    network_timeout=Item.NetworkTimeout,
    network_disconnected=Item.NetworkDisconnected,
    network_server_down=Item.NetworkServerDown,
    network_invalid_request=Item.NetworkInvalidRequest,
    server_failed=Item.ServerFailed,
    server_bad_content=Item.ServerBadContent,
    server_unauthorized=Item.ServerUnauthorized,
    server_cert_problem=Item.ServerCertProblem,
    server_forbidden=Item.ServerForbidden,
    server_unreachable=Item.ServerUnreachable,
    user_canceled=Item.UserCanceled,
)

DownloadInterruptReasonStr = Literal[
    "none",
    "file_failed",
    "file_access_denied",
    "file_no_space",
    "filename_too_long",
    "file_too_large",
    "file_virus_infected",
    "file_transient_error",
    "file_blocked",
    "file_security_check_failed",
    "file_too_short",
    "file_hash_mismatch",
    "network_fialed",
    "network_timeout",
    "network_disconnected",
    "network_server_down",
    "network_invalid_request",
    "server_failed",
    "server_bad_content",
    "server_unauthorized",
    "server_cert_problem",
    "server_forbidden",
    "server_unreachable",
    "user_canceled",
]

DOWNLOAD_STATE = bidict(
    requested=QtWebEngineWidgets.QWebEngineDownloadItem.DownloadRequested,
    in_progress=QtWebEngineWidgets.QWebEngineDownloadItem.DownloadInProgress,
    completed=QtWebEngineWidgets.QWebEngineDownloadItem.DownloadCompleted,
    cancelled=QtWebEngineWidgets.QWebEngineDownloadItem.DownloadCancelled,
    interrupted=QtWebEngineWidgets.QWebEngineDownloadItem.DownloadInterrupted,
)

DownloadStateStr = Literal[
    "requested", "in_progress", "completed", "cancelled", "interrupted"
]

SAVE_PAGE_FORMAT = bidict(
    unknown=QtWebEngineWidgets.QWebEngineDownloadItem.UnknownSaveFormat,
    single_html=QtWebEngineWidgets.QWebEngineDownloadItem.SingleHtmlSaveFormat,
    complete_html=QtWebEngineWidgets.QWebEngineDownloadItem.CompleteHtmlSaveFormat,
    mime_html=QtWebEngineWidgets.QWebEngineDownloadItem.MimeHtmlSaveFormat,
)

SavePageFormatStr = Literal["unknown", "single_html", "complete_html", "mime_html"]

# QtWebEngineWidgets.QWebEngineDownloadItem.__bases__ = (core.Object,)


class WebEngineDownloadItem:
    def __init__(self, item: QtWebEngineWidgets.QWebEngineDownloadItem):
        self.item = item

    def __getattr__(self, val):
        return getattr(self.item, val)

    def get_interrupt_reason(self) -> DownloadInterruptReasonStr:
        return DOWNLOAD_INTERRUPT_REASONS.inverse[self.item.interruptReason()]

    def get_state(self) -> DownloadStateStr:
        return DOWNLOAD_STATE.inverse[self.item.state()]

    def set_save_page_format(self, fmt: SavePageFormatStr):
        """Set the save page format.

        Args:
            fmt: save page format for the layout

        Raises:
            InvalidParamError: save page format does not exist
        """
        if fmt not in SAVE_PAGE_FORMAT:
            raise InvalidParamError(fmt, SAVE_PAGE_FORMAT)
        self.item.setSavePageFormat(SAVE_PAGE_FORMAT[fmt])

    def get_save_page_format(self) -> SavePageFormatStr:
        """Return current save page format.

        Returns:
            Save page format
        """
        return SAVE_PAGE_FORMAT.inverse[self.item.savePageFormat()]


# if __name__ == "__main__":
#     from prettyqt import widgets

# app = widgets.app()
# item = WebEngineDownloadItem()
# app.main_loop()



from __future__ import annotations

import datetime

from prettyqt import charts
from prettyqt.qt.QtCharts import QtCharts


QtCharts.QDateTimeAxis.__bases__ = (charts.AbstractAxis,)


class DateTimeAxis(QtCharts.QDateTimeAxis):
    def get_min(self) -> datetime.datetime:
        return self.min().toPython()

    def get_max(self) -> datetime.datetime:
        return self.max().toPython()



"""eventfilters module.

Contains custom EventFilter classes
"""

from .hovericoneventfilter import HoverIconEventFilter
from .animatedtooltipeventfilter import AnimatedToolTipEventFilter


__all__ = ["HoverIconEventFilter", "AnimatedToolTipEventFilter"]



from __future__ import annotations

import contextlib
from typing import Literal

from prettyqt import gui
from prettyqt.qt import QtCore, QtGui, QtQuick
from prettyqt.utils import InvalidParamError, bidict


CREATE_TEXTURE_OPTION = bidict(
    has_alpha_channel=QtQuick.QQuickWindow.TextureHasAlphaChannel,
    has_mipmaps=QtQuick.QQuickWindow.TextureHasMipmaps,
    owns_gl_texture=QtQuick.QQuickWindow.TextureOwnsGLTexture,
    can_use_atlas=QtQuick.QQuickWindow.TextureCanUseAtlas,
    is_opaque=QtQuick.QQuickWindow.TextureIsOpaque,
)

CreateTextureOptionStr = Literal[
    "has_alpha_channel",
    "has_mipmaps",
    "owns_gl_texture",
    "can_use_atlas",
    "is_opaque",
]

RENDER_STAGE = bidict(
    before_synchronizing=QtQuick.QQuickWindow.BeforeSynchronizingStage,
    after_synchronizing=QtQuick.QQuickWindow.AfterSynchronizingStage,
    before_rendering=QtQuick.QQuickWindow.BeforeRenderingStage,
    after_rendering=QtQuick.QQuickWindow.AfterRenderingStage,
    after_swap=QtQuick.QQuickWindow.AfterSwapStage,
    no_stage=QtQuick.QQuickWindow.NoStage,
)

RenderStageStr = Literal[
    "before_synchronizing",
    "after_synchronizing",
    "before_rendering",
    "after_rendering",
    "after_swap",
    "no_stage",
]

TEXT_RENDER_TYPE = bidict(
    qt_text=QtQuick.QQuickWindow.QtTextRendering,
    native_text=QtQuick.QQuickWindow.NativeTextRendering,
)

TextRenderTypeStr = Literal["qt_text", "native_text"]


QtQuick.QQuickWindow.__bases__ = (gui.Window,)


class QuickWindow(QtQuick.QQuickWindow):
    def create_texture_from_image(self, image: QtGui.QImage, **kwargs):
        flag = self.CreateTextureOptions()
        for key, val in kwargs.items():
            if val is True:
                v = CREATE_TEXTURE_OPTION[key]
                flag |= v
        return self.createTextureFromImage(image, flag)

    def grab_window(self) -> gui.Image:
        return gui.Image(self.grabWindow())

    def get_color(self) -> gui.Color:
        return gui.Color(self.color())

    @staticmethod
    def set_text_render_type(typ: TextRenderTypeStr):
        """Set the default render type of text-like elements in Qt Quick.

        Note: setting the render type will only affect elements created afterwards;
        the render type of existing elements will not be modified.

        Args:
            typ: text render type to use

        Raises:
            InvalidParamError: text render type does not exist
        """
        if typ not in TEXT_RENDER_TYPE:
            raise InvalidParamError(typ, TEXT_RENDER_TYPE)
        QuickWindow.setTextRenderType(TEXT_RENDER_TYPE[typ])

    @staticmethod
    def get_text_render_type() -> TextRenderTypeStr:
        """Return the render type of text-like elements in Qt Quick.

        Returns:
            text render type
        """
        return TEXT_RENDER_TYPE.inverse[QuickWindow.textRenderType()]

    @contextlib.contextmanager
    def external_commands(self):
        self.beginExternalCommands()
        yield self
        self.endExternalCommands()

    def schedule_render_job(self, job: QtCore.QRunnable, render_stage: RenderStageStr):
        if render_stage not in RENDER_STAGE:
            raise InvalidParamError(render_stage, RENDER_STAGE)
        self.scheduleRenderJob(job, RENDER_STAGE[render_stage])


if __name__ == "__main__":
    app = gui.GuiApplication([])
    wnd = QuickWindow()
    img = QtGui.QImage()
    texture = wnd.create_texture_from_image(img)



from __future__ import annotations

from prettyqt import gui, widgets
from prettyqt.qt import QtWidgets


QtWidgets.QGraphicsColorizeEffect.__bases__ = (widgets.GraphicsEffect,)


class GraphicsColorizeEffect(QtWidgets.QGraphicsColorizeEffect):
    def serialize_fields(self):
        return dict(strength=self.strength(), color=gui.Color(self.color()))

    def __setstate__(self, state):
        super().__setstate__(state)
        self.setStrength(state["strength"])
        self.setColor(state["color"])



from __future__ import annotations

from prettyqt import core
from prettyqt.qt import QtWebEngineCore


QtWebEngineCore.QWebEngineUrlSchemeHandler.__bases__ = (core.Object,)


class WebEngineUrlSchemeHandler(QtWebEngineCore.QWebEngineUrlSchemeHandler):
    pass


if __name__ == "__main__":
    item = WebEngineUrlSchemeHandler()



from __future__ import annotations

from typing import Optional

from prettyqt import core, gui, widgets
from prettyqt.qt import QtCore, QtGui, QtWidgets


PAINTING_SCALE_FACTOR = 20
DIAMOND_POLYGON = gui.PolygonF.create_diamond()
STAR_POLYGON = gui.PolygonF.create_star()


class StarRating:
    """Handle the actual painting of the stars themselves."""

    def __init__(self, star_count: int = 1, max_stars: int = 5):
        self.star_count = star_count
        self.max_stars = max_stars

    def sizeHint(self):
        return PAINTING_SCALE_FACTOR * core.Size(self.max_stars, 1)

    def paint(
        self,
        painter: QtGui.QPainter,
        rect: QtCore.QRect,
        palette: QtGui.QPalette,
        is_editable: bool = False,
    ):
        """Paint the stars (and/or diamonds if we're in editing mode)."""
        painter.save()
        painter.setRenderHint(painter.Antialiasing, True)
        painter.setPen(QtCore.Qt.NoPen)
        painter.setBrush(palette.highlight() if is_editable else palette.windowText())
        y_offset = (rect.height() - PAINTING_SCALE_FACTOR) / 2
        painter.translate(rect.x(), rect.y() + y_offset)
        painter.scale(PAINTING_SCALE_FACTOR, PAINTING_SCALE_FACTOR)
        for i in range(self.max_stars):
            if i < self.star_count:
                painter.drawPolygon(STAR_POLYGON, QtCore.Qt.WindingFill)
            elif is_editable:
                painter.drawPolygon(DIAMOND_POLYGON, QtCore.Qt.WindingFill)
            else:
                break
            painter.translate(1.0, 0.0)
        painter.restore()


class StarEditor(widgets.Widget):
    """The custom editor for editing StarRatings."""

    # A signal to tell the delegate when we've finished editing.
    editing_finished = core.Signal()

    def __init__(self, parent: Optional[QtWidgets.QWidget] = None):
        """Initialize the editor object, making sure we can watch mouse events."""
        super().__init__(parent)

        self.setMouseTracking(True)
        self.setAutoFillBackground(True)
        self.star_rating = StarRating()

    def sizeHint(self):
        """Tell the caller how big we are."""
        return self.star_rating.sizeHint()

    def paintEvent(self, event):
        """Paint the editor, offloading the work to the StarRating class."""
        painter = gui.Painter(self)
        self.star_rating.paint(painter, self.rect(), self.palette(), is_editable=True)

    def mouseMoveEvent(self, event):
        """Update stars on mouse move."""
        star = self.star_at_position(event.x())

        if star != -1:
            self.star_rating.star_count = star
            self.update()

    def mouseReleaseEvent(self, event):
        """Once star rating was clicked, tell the delegate we're done editing."""
        self.editing_finished.emit()

    def star_at_position(self, x: int) -> int:
        """Calculate which star the user's mouse cursor is currently hovering over."""
        val = x // (self.star_rating.sizeHint().width() // self.star_rating.max_stars) + 1
        if not 0 < val <= self.star_rating.max_stars:
            return -1
        return val

    def set_star_rating(self, rating: int):
        self.star_rating.star_count = rating


class StarDelegate(widgets.StyledItemDelegate):
    """A delegate class that allows us to render our star ratings."""

    def paint(self, painter, option, index):
        star_rating = StarRating(index.data())

        # If the row is currently selected, we need to make sure we
        # paint the background accordingly.
        if option.state & widgets.Style.State_Selected:
            # The original C++ example used option.palette.foreground() to
            # get the brush for painting, but there are a couple of
            # problems with that:
            #   - foreground() is obsolete now, use windowText() instead
            #   - more importantly, windowText() just returns a brush
            #     containing a flat color, where sometimes the style
            #     would have a nice subtle gradient or something.
            # Here we just use the brush of the painter object that's
            # passed in to us, which keeps the row highlighting nice
            # and consistent.
            painter.fillRect(option.rect, painter.brush())

        # Now that we've painted the background, call star_rating.paint()
        # to paint the stars.
        star_rating.paint(painter, option.rect, option.palette)

    def sizeHint(self, option, index):
        """Return the size needed to display the item in a QSize object."""
        star_rating = StarRating(index.data())
        return star_rating.sizeHint()

    # The next 4 methods handle the custom editing that we need to do.
    # If this were just a display delegate, paint() and sizeHint() would
    # be all we needed.

    def createEditor(self, parent, option, index):
        """Create and return the StarEditor object we'll use to edit the StarRating."""
        editor = StarEditor(parent)
        editor.editing_finished.connect(self.commitAndCloseEditor)
        return editor

    def setEditorData(self, editor, index):
        """Set the data to be displayed and edited by our custom editor."""
        editor.set_star_rating(index.data())

    def setModelData(self, editor, model, index):
        """Get the data from our custom editor and stuffs it into the model."""
        model.setData(index, editor.star_rating.star_count)

    def commitAndCloseEditor(self):
        editor = self.sender()

        # The commitData signal must be emitted when we've finished editing
        # and need to write our changed back to the model.
        self.commitData.emit(editor)
        self.closeEditor.emit(editor, self.NoHint)


if __name__ == "__main__":
    """Run the application."""
    from prettyqt import widgets

    app = widgets.app()

    # Create and populate the tableWidget
    table_widget = widgets.TableWidget(1, 2)
    table_widget.set_delegate(StarDelegate(), column=1)
    table_widget.setEditTriggers(
        widgets.AbstractItemView.DoubleClicked | widgets.AbstractItemView.SelectedClicked
    )
    table_widget.set_selection_behaviour("rows")
    table_widget.setHorizontalHeaderLabels(["Title", "Rating"])

    item_1 = widgets.TableWidgetItem("Test1")
    # item_1.setData(0, 3)
    item_2 = widgets.TableWidgetItem()
    item_2.setData(0, 3)
    table_widget[0, 0] = item_1
    table_widget[0, 1] = item_2

    table_widget.resizeColumnsToContents()
    table_widget.resize(500, 300)
    table_widget.show()

    app.main_loop()



from __future__ import annotations

from typing import Iterator, Literal, Tuple

from prettyqt import charts
from prettyqt.qt.QtCharts import QtCharts
from prettyqt.utils import InvalidParamError, bidict


LABELS_POSITIONS = bidict(
    center=QtCharts.QCategoryAxis.AxisLabelsPositionCenter,
    on_value=QtCharts.QCategoryAxis.AxisLabelsPositionOnValue,
)

LabelsPositionStr = Literal["center", "on_value"]

QtCharts.QCategoryAxis.__bases__ = (charts.ValueAxis,)


class CategoryAxis(QtCharts.QCategoryAxis):
    def __delitem__(self, index: str):
        self.remove(index)

    def __getitem__(self, label: str) -> str:
        return self.categoriesLabels()[label]

    def __setitem__(self, index: str, value: str):
        self.replaceLabel(index, value)

    def __iter__(self) -> Iterator[str]:
        return iter(self.categoriesLabels())

    def __add__(self, other: Tuple[str, int]) -> CategoryAxis:
        self.append(*other)
        return self

    def __len__(self):
        # needed for PySide2
        return self.count()

    def set_labels_position(self, position: LabelsPositionStr):
        """Set the labels position.

        Args:
            position: labels position

        Raises:
            InvalidParamError: labels position does not exist
        """
        if position not in LABELS_POSITIONS:
            raise InvalidParamError(position, LABELS_POSITIONS)
        self.setLabelsPosition(LABELS_POSITIONS[position])

    def get_labels_position(self) -> LabelsPositionStr:
        """Return current labels position.

        Returns:
            labels position
        """
        return LABELS_POSITIONS.inverse[self.labelsPosition()]


if __name__ == "__main__":
    axis = CategoryAxis()



from __future__ import annotations

from typing import Literal

from prettyqt.qt import QtGui
from prettyqt.utils import bidict


PAGE_SIZE_ID = bidict(
    a0=QtGui.QPageSize.A0,
    a1=QtGui.QPageSize.A1,
    a2=QtGui.QPageSize.A2,
    a3=QtGui.QPageSize.A3,
    a4=QtGui.QPageSize.A4,
    a5=QtGui.QPageSize.A5,
    a6=QtGui.QPageSize.A6,
    a7=QtGui.QPageSize.A7,
    a8=QtGui.QPageSize.A8,
    a9=QtGui.QPageSize.A9,
    b0=QtGui.QPageSize.B0,
    b1=QtGui.QPageSize.B1,
    b2=QtGui.QPageSize.B2,
    b3=QtGui.QPageSize.B3,
    b4=QtGui.QPageSize.B4,
    b5=QtGui.QPageSize.B5,
    b6=QtGui.QPageSize.B6,
    b7=QtGui.QPageSize.B7,
    b8=QtGui.QPageSize.B8,
    b9=QtGui.QPageSize.B9,
    b10=QtGui.QPageSize.B10,
    c5e=QtGui.QPageSize.C5E,
    comm_10e=QtGui.QPageSize.Comm10E,
    dle=QtGui.QPageSize.DLE,
    executive=QtGui.QPageSize.Executive,
    folio=QtGui.QPageSize.Folio,
    ledger=QtGui.QPageSize.Ledger,
    legal=QtGui.QPageSize.Legal,
    letter=QtGui.QPageSize.Letter,
    tabloid=QtGui.QPageSize.Tabloid,
    custom=QtGui.QPageSize.Custom,
    a10=QtGui.QPageSize.A10,
    a3_extra=QtGui.QPageSize.A3Extra,
    a4_extra=QtGui.QPageSize.A4Extra,
    a4_plus=QtGui.QPageSize.A4Plus,
    a4_small=QtGui.QPageSize.A4Small,
    a5_extra=QtGui.QPageSize.A5Extra,
    b5_extra=QtGui.QPageSize.B5Extra,
    jis_b0=QtGui.QPageSize.JisB0,
    jis_b1=QtGui.QPageSize.JisB1,
    jis_b2=QtGui.QPageSize.JisB2,
    jis_b3=QtGui.QPageSize.JisB3,
    jis_b4=QtGui.QPageSize.JisB4,
    jis_b5=QtGui.QPageSize.JisB5,
    jis_b6=QtGui.QPageSize.JisB6,
    jis_b7=QtGui.QPageSize.JisB7,
    jis_b8=QtGui.QPageSize.JisB8,
    jis_b9=QtGui.QPageSize.JisB9,
    jis_b10=QtGui.QPageSize.JisB10,
    # c5e=QtGui.QPageSize.AnsiA, Letter
    # c5e=QtGui.QPageSize.AnsiB, Ledger
    ansi_c=QtGui.QPageSize.AnsiC,
    ansi_d=QtGui.QPageSize.AnsiD,
    ansi_e=QtGui.QPageSize.AnsiE,
    legal_extra=QtGui.QPageSize.LegalExtra,
    letter_extra=QtGui.QPageSize.LetterExtra,
    letter_plus=QtGui.QPageSize.LetterPlus,
    letter_small=QtGui.QPageSize.LetterSmall,
    tabloid_extra=QtGui.QPageSize.TabloidExtra,
    arch_a=QtGui.QPageSize.ArchA,
    arch_b=QtGui.QPageSize.ArchB,
    arch_c=QtGui.QPageSize.ArchC,
    arch_d=QtGui.QPageSize.ArchD,
    arch_e=QtGui.QPageSize.ArchE,
    imperial_7x9=QtGui.QPageSize.Imperial7x9,
    imperial_8x10=QtGui.QPageSize.Imperial8x10,
    imperial_9x11=QtGui.QPageSize.Imperial9x11,
    imperial_9x12=QtGui.QPageSize.Imperial9x12,
    imperial_10x11=QtGui.QPageSize.Imperial10x11,
    imperial_10x13=QtGui.QPageSize.Imperial10x13,
    imperial_10x14=QtGui.QPageSize.Imperial10x14,
    imperial_12x11=QtGui.QPageSize.Imperial12x11,
    imperial_15x11=QtGui.QPageSize.Imperial15x11,
    executive_standard=QtGui.QPageSize.ExecutiveStandard,
    note=QtGui.QPageSize.Note,
    quarto=QtGui.QPageSize.Quarto,
    statement=QtGui.QPageSize.Statement,
    super_a=QtGui.QPageSize.SuperA,
    super_b=QtGui.QPageSize.SuperB,
    postcard=QtGui.QPageSize.Postcard,
    double_postcard=QtGui.QPageSize.DoublePostcard,
    prc_16_k=QtGui.QPageSize.Prc16K,
    prc_32_k=QtGui.QPageSize.Prc32K,
    prc_32_k_big=QtGui.QPageSize.Prc32KBig,
    fan_fold_us=QtGui.QPageSize.FanFoldUS,
    fan_fold_german=QtGui.QPageSize.FanFoldGerman,
    fan_fold_german_legal=QtGui.QPageSize.FanFoldGermanLegal,
    envelope_b4=QtGui.QPageSize.EnvelopeB4,
    envelope_b5=QtGui.QPageSize.EnvelopeB5,
    envelope_b6=QtGui.QPageSize.EnvelopeB6,
    envelope_c0=QtGui.QPageSize.EnvelopeC0,
    envelope_c1=QtGui.QPageSize.EnvelopeC1,
    envelope_c2=QtGui.QPageSize.EnvelopeC2,
    envelope_c3=QtGui.QPageSize.EnvelopeC3,
    envelope_c4=QtGui.QPageSize.EnvelopeC4,
    # envelope_c5=QtGui.QPageSize.EnvelopeC5, c5e
    envelope_c6=QtGui.QPageSize.EnvelopeC6,
    envelope_c65=QtGui.QPageSize.EnvelopeC65,
    envelope_c7=QtGui.QPageSize.EnvelopeC7,
    # envelope_dl=QtGui.QPageSize.EnvelopeDL, dle
    envelope_9=QtGui.QPageSize.Envelope9,
    # envelope_10=QtGui.QPageSize.Envelope10, comm10e
    envelope_11=QtGui.QPageSize.Envelope11,
    envelope_12=QtGui.QPageSize.Envelope12,
    envelope_14=QtGui.QPageSize.Envelope14,
    envelope_monarch=QtGui.QPageSize.EnvelopeMonarch,
    envelope_personal=QtGui.QPageSize.EnvelopePersonal,
    envelope_chou_3=QtGui.QPageSize.EnvelopeChou3,
    envelope_chou_4=QtGui.QPageSize.EnvelopeChou4,
    envelope_invite=QtGui.QPageSize.EnvelopeInvite,
    envelope_italian=QtGui.QPageSize.EnvelopeItalian,
    envelope_kaku_2=QtGui.QPageSize.EnvelopeKaku2,
    envelope_kaku_3=QtGui.QPageSize.EnvelopeKaku3,
    envelope_prc_1=QtGui.QPageSize.EnvelopePrc1,
    envelope_prc_2=QtGui.QPageSize.EnvelopePrc2,
    envelope_prc_3=QtGui.QPageSize.EnvelopePrc3,
    envelope_prc_4=QtGui.QPageSize.EnvelopePrc4,
    envelope_prc_5=QtGui.QPageSize.EnvelopePrc5,
    envelope_prc_6=QtGui.QPageSize.EnvelopePrc6,
    envelope_prc_7=QtGui.QPageSize.EnvelopePrc7,
    envelope_prc_8=QtGui.QPageSize.EnvelopePrc8,
    envelope_prc_9=QtGui.QPageSize.EnvelopePrc9,
    envelope_prc_10=QtGui.QPageSize.EnvelopePrc10,
    envelope_you_4=QtGui.QPageSize.EnvelopeYou4,
    # last_page_size=QtGui.QPageSize.LastPageSize, envelope_you_4
)

SIZE_MATCH_POLICY = bidict(
    fuzzy=QtGui.QPageSize.FuzzyMatch,
    fuzzy_orientation=QtGui.QPageSize.FuzzyOrientationMatch,
    exact=QtGui.QPageSize.ExactMatch,
)

UNITS = bidict(
    millimeter=QtGui.QPageSize.Millimeter,
    point=QtGui.QPageSize.Point,
    inch=QtGui.QPageSize.Inch,
    pica=QtGui.QPageSize.Pica,
    didot=QtGui.QPageSize.Didot,
    cicero=QtGui.QPageSize.Cicero,
)

UnitStr = Literal["millimeter", "point", "inch", "pica", "didot", "cicero"]


class PageSize(QtGui.QPageSize):
    def __repr__(self):
        return f"{type(self).__name__}({self.get_id()!r})"

    def __reduce__(self):
        return type(self), (self.id(),)

    def get_definition_units(self) -> UnitStr:
        """Get the definition unit.

        Returns:
            unit
        """
        units = self.definitionUnits()
        if units == -1:
            raise ValueError("Invalid page size")
        return UNITS.inverse[units]

    def get_id(self) -> str:
        """Get the standard page size id.

        Returns:
            page size id
        """
        return PAGE_SIZE_ID.inverse[self.id()]


if __name__ == "__main__":
    size = PageSize()
    print(repr(size))



"""Module containing custom animation classes."""

from .bounceanimation import BounceAnimation
from .fadeinanimation import FadeInAnimation
from .slideanimation import SlideAnimation


__all__ = ["BounceAnimation", "SlideAnimation", "FadeInAnimation"]



from __future__ import annotations

from typing import List

from prettyqt import core
from prettyqt.qt import QtCore, QtGui, QtWidgets


TYPES = {
    bool: 1,
    int: 2,
    str: 10,
    float: 38,
    QtGui.QColor: 67,
    QtGui.QCursor: 74,
    QtCore.QDate: 14,
    QtCore.QSize: 21,
    QtCore.QTime: 15,
    list: 9,
    QtGui.QPolygon: 71,
    QtGui.QPolygonF: 86,
    QtGui.QColor: 67,
    QtGui.QColorSpace: 87,
    QtCore.QSizeF: 22,
    QtCore.QRectF: 20,
    QtCore.QLine: 23,
    QtGui.QTextLength: 77,
    dict: 8,
    QtGui.QIcon: 69,
    QtGui.QPen: 76,
    QtCore.QLineF: 24,
    QtGui.QTextFormat: 78,
    QtCore.QRect: 19,
    QtCore.QPoint: 25,
    QtCore.QUrl: 17,
    QtCore.QRegularExpression: 44,
    QtCore.QDateTime: 16,
    QtCore.QPointF: 26,
    QtGui.QPalette: 68,
    QtGui.QFont: 64,
    QtGui.QBrush: 66,
    QtGui.QRegion: 72,
    QtGui.QImage: 70,
    QtGui.QKeySequence: 75,
    QtWidgets.QSizePolicy: 121,
    QtGui.QPixmap: 65,
    QtCore.QLocale: 18,
    QtGui.QBitmap: 73,
    QtGui.QMatrix4x4: 81,
    QtGui.QVector2D: 82,
    QtGui.QVector3D: 83,
    QtGui.QVector4D: 84,
    QtGui.QQuaternion: 85,
    QtCore.QEasingCurve: 29,
    QtCore.QJsonValue: 45,
    QtCore.QJsonDocument: 48,
    QtCore.QModelIndex: 42,
    QtCore.QPersistentModelIndex: 50,
    QtCore.QUuid: 30,
    "user": 1024,
}


class MetaObject:
    def __init__(self, metaobject: QtCore.QMetaObject):
        self.item = metaobject

    def get_enumerators(self) -> List[dict]:
        enums = []
        for i in range(self.item.enumeratorCount()):
            enumerator = self.item.enumerator(i)
            dct = dict(
                name=enumerator.name(),
                enum_name=enumerator.enumName(),
                is_flag=enumerator.isFlag(),
                is_scoped=enumerator.isScoped(),
                key_count=enumerator.keyCount(),
            )
            enums.append(dct)
        return enums

    def get_methods(self) -> List[core.MetaMethod]:
        return [
            core.MetaMethod(self.item.method(i)) for i in range(self.item.methodCount())
        ]

    def get_constructors(self) -> List[core.MetaMethod]:
        return [
            core.MetaMethod(self.item.constructor(i))
            for i in range(self.item.constructorCount())
        ]



from __future__ import annotations

import os
import pathlib
from typing import List, Union

from prettyqt import core
from prettyqt.qt import QtCore


QtCore.QFileSystemWatcher.__bases__ = (core.Object,)


class FileSystemWatcher(QtCore.QFileSystemWatcher):
    def __repr__(self):
        paths = [str(p) for p in self.get_paths()]
        return f"{type(self).__name__}({paths})"

    def get_directories(self) -> List[pathlib.Path]:
        return [pathlib.Path(p) for p in self.directories()]

    def get_files(self) -> List[pathlib.Path]:
        return [pathlib.Path(p) for p in self.files()]

    def get_paths(self) -> List[pathlib.Path]:
        return self.get_directories() + self.get_files()

    def add_path(self, path: Union[str, os.PathLike]) -> bool:
        return self.addPath(os.fspath(path))

    def add_paths(self, paths: List[Union[str, os.PathLike]]):
        self.addPaths([os.fspath(p) for p in paths])



from __future__ import annotations

import os
from typing import Iterator, Literal, Optional, Union

from prettyqt import core, multimedia
from prettyqt.qt import QtMultimedia
from prettyqt.utils import InvalidParamError, bidict


PLAYBACK_MODE = bidict(
    item_once=QtMultimedia.QMediaPlaylist.CurrentItemOnce,
    item_in_loop=QtMultimedia.QMediaPlaylist.CurrentItemInLoop,
    sequential=QtMultimedia.QMediaPlaylist.Sequential,
    loop=QtMultimedia.QMediaPlaylist.Loop,
    random=QtMultimedia.QMediaPlaylist.Random,
)

PlaybackModeStr = Literal["item_once", "item_in_loop", "sequential", "loop", "random"]

QtMultimedia.QMediaPlaylist.__bases__ = (core.Object, multimedia.MediaBindableInterface)


class MediaPlaylist(QtMultimedia.QMediaPlaylist):
    playback_mode_changed = core.Signal(str)

    def __len__(self) -> int:
        return self.mediaCount()

    def __getitem__(self, item: int) -> multimedia.MediaContent:
        return multimedia.MediaContent(self.media(item))

    def __delitem__(self, item: int):
        self.removeMedia(item)

    def __iter__(self) -> Iterator[multimedia.MediaContent]:
        return iter(self[i] for i in range(self.mediaCount()))

    # def serialize(self) -> Dict[str, Any]:
    #     return dict(current_index=self.currentIndex(),
    #                 playback_mode=self.get_playback_mode(),
    #                 items=list(self))

    def add_media(
        self, media: Union[os.PathLike, str], pos: Optional[int] = None
    ) -> bool:
        url = core.Url(os.fspath(media))
        mediacontent = multimedia.MediaContent(url)
        if pos is None:
            return self.addMedia(mediacontent)
        else:
            return self.insertMedia(pos, mediacontent)

    def get_media_url(self, index: int) -> core.Url:
        url = self.media(index).request().url()
        return core.Url(url)

    def set_playback_mode(self, mode: PlaybackModeStr):
        """Set playback mode for given item view.

        Args:
            mode: playback mode to use

        Raises:
            InvalidParamError: mode does not exist
        """
        if mode not in PLAYBACK_MODE:
            raise InvalidParamError(mode, PLAYBACK_MODE)
        self.setPlaybackMode(PLAYBACK_MODE[mode])

    def get_playback_mode(self) -> PlaybackModeStr:
        """Return current playback mode.

        Returns:
            playback mode
        """
        return PLAYBACK_MODE.inverse[self.playbackMode()]

    # def on_playback_mode_changed(self, mode: int):
    #     self.playback_mode_changed.emit(PLAYBACK_MODE.inverse[mode])



"""Module containing helper functions."""

from __future__ import annotations

import collections
import logging
import sys
import traceback

from prettyqt import widgets
from prettyqt.qt import QtCore


logger = logging.getLogger(__name__)

LOG_MAP = {
    QtCore.QtMsgType.QtInfoMsg: 20,
    QtCore.QtMsgType.QtWarningMsg: 30,
    QtCore.QtMsgType.QtCriticalMsg: 40,
    QtCore.QtMsgType.QtFatalMsg: 50,
}


def qt_message_handler(mode: QtCore.QtMsgType, context, message: str):
    level = LOG_MAP.get(mode, 20)
    logger.log(level, f"{message} ({context.file}:{context.line}, {context.file})")


def install_exceptionhook(debug: bool = False):
    def handleException(exc_type, exc_value, exc_traceback):
        """Causes the application to quit in case of an unhandled exception.

        Shows an error dialog before quitting when not in debugging mode.
        """
        logger.critical(
            f"Bug: uncaught {exc_type.__name__}",
            exc_info=(exc_type, exc_value, exc_traceback),
        )
        if debug:
            sys.exit(1)
        else:
            from prettyqt import widgets

            # Constructing a QApplication in case this hasn't been done yet.
            _ = widgets.app()
            lst = traceback.format_exception(exc_type, exc_value, exc_traceback)
            msgBox = widgets.MessageBox(
                icon="warning",
                text=f"Bug: uncaught {exc_type.__name__}",
                informative_text=str(exc_value),
                details="".join(lst),
            )
            msgBox.exec_()
            sys.exit(1)

    sys.excepthook = handleException


def count_objects():
    win = widgets.Application.get_mainwindow()
    objects = win.findChildren(QtCore.QObject)
    counter = collections.Counter([type(o) for o in objects])
    logger.info(counter)



from __future__ import annotations

from prettyqt import constants, core, gui
from prettyqt.qt import QtWidgets
from prettyqt.utils import InvalidParamError


QtWidgets.QShortcut.__bases__ = (core.Object,)


class Shortcut(QtWidgets.QShortcut):
    def __str__(self):
        return self.key().toString()

    def __int__(self):
        return self.id()

    def serialize_field(self):
        return dict(
            auto_repeat=self.autoRepeat(),
            context=self.get_context(),
            enabled=self.isEnabled(),
            key=self.get_key(),
            whats_this=self.whatsThis(),
        )

    def set_context(self, context: constants.ContextStr):
        """Set shortcut context.

        Args:
            context: shortcut context

        Raises:
            InvalidParamError: shortcut context does not exist
        """
        if context not in constants.CONTEXT:
            raise InvalidParamError(context, constants.CONTEXT)
        self.setContext(constants.CONTEXT[context])

    def get_context(self) -> constants.ContextStr:
        """Return shortcut context.

        Returns:
            shortcut context
        """
        return constants.CONTEXT.inverse[self.context()]

    def get_key(self) -> gui.KeySequence:
        """Return the shortcut's key sequence.

        Returns:
            Key sequence
        """
        return gui.KeySequence(self.key().toString(), gui.KeySequence.PortableText)



from __future__ import annotations

from prettyqt import core
from prettyqt.qt import QtCore


QtCore.QAbstractState.__bases__ = (core.Object,)


class AbstractState(QtCore.QAbstractState):
    def serialize_fields(self):
        return dict(active=self.active())


if __name__ == "__main__":
    state = AbstractState()



from __future__ import annotations

from typing import Literal

from prettyqt import charts
from prettyqt.qt.QtCharts import QtCharts
from prettyqt.utils import InvalidParamError, bidict


POLAR_ORIENTATIONS = bidict(
    radial=QtCharts.QPolarChart.PolarOrientationRadial,
    angular=QtCharts.QPolarChart.PolarOrientationAngular,
)

PolarOrientationStr = Literal["radial", "angular"]

QtCharts.QPolarChart.__bases__ = (charts.Chart,)


class PolarChart(QtCharts.QPolarChart):
    def add_axis(self, axis: QtCharts.QAbstractAxis, orientation: PolarOrientationStr):
        if orientation not in POLAR_ORIENTATIONS:
            raise InvalidParamError(orientation, POLAR_ORIENTATIONS)
        self.addAxis(axis, POLAR_ORIENTATIONS[orientation])



from __future__ import annotations

from prettyqt import widgets
from prettyqt.qt import QtWidgets


QtWidgets.QGraphicsProxyWidget.__bases__ = (widgets.GraphicsWidget,)


class GraphicsProxyWidget(QtWidgets.QGraphicsProxyWidget):

    pass



from __future__ import annotations

from prettyqt import core
from prettyqt.qt import QtWidgets


QtWidgets.QAbstractItemDelegate.__bases__ = (core.Object,)


class AbstractItemDelegate(QtWidgets.QAbstractItemDelegate):
    pass



from __future__ import annotations

from prettyqt.qt import QtCore


class DirIterator(QtCore.QDirIterator):
    def __iter__(self):
        return self

    def __next__(self):
        if self.hasNext():
            return self.next()
        raise StopIteration



from __future__ import annotations

import os
from typing import Literal, Union

from prettyqt import core
from prettyqt.qt import QtCore, QtMultimedia
from prettyqt.utils import bidict


STATUS = bidict(
    null=QtMultimedia.QSoundEffect.Null,
    loading=QtMultimedia.QSoundEffect.Loading,
    ready=QtMultimedia.QSoundEffect.Ready,
    error=QtMultimedia.QSoundEffect.Error,
)

StatusStr = Literal["null", "loading", "ready", "error"]

QtMultimedia.QSoundEffect.__bases__ = (core.Object,)


class SoundEffect(QtMultimedia.QSoundEffect):
    def get_status(self) -> StatusStr:
        return STATUS.inverse[self.status()]

    def set_source(self, source: Union[str, os.PathLike, QtCore.QUrl]):
        if not isinstance(source, QtCore.QUrl):
            source = core.Url.from_user_input(os.fspath(source))
        self.setSource(source)

    def get_source(self) -> core.Url:
        return core.Url(self.source())

    def set_loop_count(self, count: Union[Literal["inf"], int]):
        if count == "inf":
            count = QtMultimedia.QSoundEffect.Infinite
        self.setLoopCount(count)



from __future__ import annotations

from typing import Any, Iterable, Iterator, List, Mapping, Optional, Union

from prettyqt import constants, core, iconprovider, widgets
from prettyqt.qt import QtCore, QtGui, QtWidgets
from prettyqt.utils import InvalidParamError, types


QtWidgets.QListWidget.__bases__ = (widgets.ListView,)


class NoData:
    pass


class ListWidget(QtWidgets.QListWidget):

    value_changed = core.Signal(object)

    def __init__(
        self, parent: Optional[QtWidgets.QWidget] = None, selection_mode: str = "single"
    ):
        super().__init__(parent)
        self.itemSelectionChanged.connect(self.on_index_change)
        self.set_selection_mode(selection_mode)

    def __repr__(self):
        return f"{type(self).__name__}: {self.count()} items"

    def __getitem__(self, row: int) -> QtWidgets.QListWidgetItem:
        item = self.item(row)
        if item is None:
            raise KeyError(row)
        return item

    def __delitem__(self, row: int):
        self.takeItem(row)

    def __add__(self, other: QtWidgets.QListWidgetItem):
        self.addItem(other)
        return self

    def __iter__(self) -> Iterator[QtWidgets.QListWidgetItem]:
        return iter(self.get_children())

    def __len__(self) -> int:
        return self.count()

    def __setstate__(self, state):
        self.set_selection_mode(state["selection_mode"])
        self.setSortingEnabled(state["sorting_enabled"])
        self.setCurrentRow(state["current_row"])
        for item in state["items"]:
            self.addItem(item)

    def __reduce__(self):
        return type(self), (), self.__getstate__()

    def serialize_fields(self):
        return dict(
            items=self.get_children(),
            selection_mode=self.get_selection_mode(),
            sorting_enabled=self.isSortingEnabled(),
            current_row=self.currentRow(),
        )

    def sort(self, reverse: bool = False):
        order = QtCore.Qt.DescendingOrder if reverse else QtCore.Qt.AscendingOrder
        self.sortItems(order)

    def on_index_change(self):
        data = self.get_value()
        self.value_changed.emit(data)

    def get_children(self) -> List[QtWidgets.QListWidgetItem]:
        return [self.item(row) for row in range(self.count())]

    def add_items(self, items: Union[Iterable, Mapping]):
        if isinstance(items, Mapping):
            for k, v in items.items():
                self.add(v, k)
        else:
            for i in items:
                if isinstance(i, (tuple, list)):
                    self.add(*i)
                else:
                    self.add(i)

    def add_item(
        self,
        name: str = "",
        icon: types.IconType = None,
        data: Optional[dict] = None,
        foreground: Optional[QtGui.QBrush] = None,
        background: Optional[QtGui.QBrush] = None,
        font: Optional[QtGui.QFont] = None,
        selected: bool = None,
        status_tip: Optional[str] = None,
        tool_tip: Optional[str] = None,
        whats_this: Optional[str] = None,
        # text_alignment: Optional[str] = None,
        checkstate: Optional[constants.StateStr] = None,
        flags: Optional[QtCore.Qt.ItemFlags] = None,
        size_hint: Optional[QtCore.QSize] = None,
        is_user_type: bool = False,
    ) -> widgets.ListWidgetItem:
        typ = 1 if is_user_type else 0
        item = widgets.ListWidgetItem(name, self, typ)
        if icon is not None:
            icon = iconprovider.get_icon(icon)
            item.setIcon(icon)
        if data is not None:
            for k, v in data.items():
                item.setData(k, v)
        if foreground is not None:
            item.setForeground(foreground)
        if background is not None:
            item.setBackground(background)
        if font is not None:
            item.setFont(font)
        if flags is not None:
            item.setFlags(flags)
        if selected:
            item.setSelected(selected)
        if status_tip:
            item.setStatusTip(status_tip)
        if tool_tip:
            item.setToolTip(tool_tip)
        if whats_this:
            item.setWhatsThis(whats_this)
        if size_hint is not None:
            item.setSizeHint(size_hint)
        if checkstate is not None:
            item.set_checkstate(checkstate)
        self.addItem(item)
        return item

    def add(self, label: str, data=NoData, icon: types.IconType = None):
        if data is NoData:
            data = label
        item = widgets.ListWidgetItem(label)
        item.set_icon(icon)
        item.setData(QtCore.Qt.UserRole, data)
        self.addItem(item)

    def get_value(self) -> List[Any]:
        return [i.data(QtCore.Qt.UserRole) for i in self.selectedItems()]

    def set_value(self, value):
        for i in self.get_children():
            if i.data(QtCore.Qt.UserRole) in value:
                self.setCurrentItem(i)
                break

    def scroll_to_item(
        self,
        item: QtWidgets.QListWidgetItem,
        mode: widgets.abstractitemview.ScrollHintStr = "ensure_visible",
    ):
        if mode not in widgets.abstractitemview.SCROLL_HINT:
            raise InvalidParamError(mode, widgets.abstractitemview.SCROLL_HINT)
        self.scrollToItem(item, widgets.abstractitemview.SCROLL_HINT[mode])


if __name__ == "__main__":
    app = QtWidgets.QApplication([])
    widget = ListWidget()
    widget.add("test", icon="mdi.timer")
    widget.add("test", icon="mdi.timer")
    widget.show()
    app.main_loop()



from __future__ import annotations

from typing import Optional

from prettyqt import gui
from prettyqt.qt import QtGui


QtGui.QIntValidator.__bases__ = (gui.Validator,)


class IntValidator(QtGui.QIntValidator):
    def __repr__(self):
        return f"{type(self).__name__}({self.bottom()}, {self.top()})"

    def __getstate__(self):
        return dict(bottom=self.bottom(), top=self.top())

    def __reduce__(self):
        return type(self), (self.bottom(), self.top()), None

    def __eq__(self, other: object):
        if not isinstance(other, type(self)):
            return False
        return self.bottom() == other.bottom() and self.top() == other.top()

    def set_range(self, lower: Optional[int], upper: Optional[int]):
        if lower is None:
            lower = 2147483647  # number from docs
        if upper is None:
            upper = 2147483647
        self.setRange(lower, upper)


if __name__ == "__main__":
    val = IntValidator()
    val.setRange(0, 9)



from __future__ import annotations

from prettyqt import widgets
from prettyqt.qt import QtWidgets


QtWidgets.QErrorMessage.__bases__ = (widgets.BaseDialog,)


class ErrorMessage(QtWidgets.QErrorMessage):
    pass


if __name__ == "__main__":
    app = widgets.app()
    widget = ErrorMessage()
    widget.set_icon("mdi.timer")
    widget.show()
    app.main_loop()



from __future__ import annotations

from prettyqt import widgets
from prettyqt.qt import QtWidgets


QtWidgets.QSizeGrip.__bases__ = (widgets.Widget,)


class SizeGrip(QtWidgets.QSizeGrip):
    pass


if __name__ == "__main__":
    app = widgets.app()
    w = widgets.Widget()
    widget = SizeGrip(w)
    widget.show()
    app.main_loop()



from __future__ import annotations

from prettyqt import widgets


class ProgressBarDelegate(widgets.StyledItemDelegate):
    def paint(self, painter, option, index):
        progress = index.data()
        progressBar_option = widgets.StyleOptionProgressBar()
        progressBar_option.rect = option.rect
        progressBar_option.minimum = 0
        progressBar_option.maximum = 100
        progressBar_option.progress = progress
        progressBar_option.text = f"{progress}%"
        progressBar_option.textVisible = True
        widgets.Application.style().drawControl(
            widgets.Style.CE_ProgressBar, progressBar_option, painter
        )


if __name__ == "__main__":
    """Run the application."""
    from prettyqt import widgets

    app = widgets.app()

    # Create and populate the tableWidget
    table_widget = widgets.TableWidget(1, 2)
    table_widget.set_delegate(ProgressBarDelegate(), column=1)
    table_widget.setEditTriggers(
        widgets.AbstractItemView.DoubleClicked | widgets.AbstractItemView.SelectedClicked
    )
    table_widget.set_selection_behaviour("rows")
    table_widget.setHorizontalHeaderLabels(["Title", "Progress"])
    item_1 = widgets.TableWidgetItem("Test1")
    item_2 = widgets.TableWidgetItem()
    item_2.setData(0, 50)
    table_widget[0, 0] = item_1
    table_widget[0, 1] = item_2
    table_widget.resizeColumnsToContents()
    table_widget.resize(500, 300)
    table_widget.show()

    app.main_loop()



from __future__ import annotations

from typing import List, Literal

from prettyqt import widgets
from prettyqt.qt import QtWidgets
from prettyqt.utils import InvalidParamError, bidict, helpers


BLUR_HINTS = bidict(
    performance=QtWidgets.QGraphicsBlurEffect.PerformanceHint,
    quality=QtWidgets.QGraphicsBlurEffect.QualityHint,
    animation=QtWidgets.QGraphicsBlurEffect.AnimationHint,
)

BlurHintStr = Literal["performance", "quality", "animation"]


QtWidgets.QGraphicsBlurEffect.__bases__ = (widgets.GraphicsEffect,)


class GraphicsBlurEffect(QtWidgets.QGraphicsBlurEffect):
    def serialize_fields(self):
        return dict(blur_radius=self.blurRadius(), blur_hints=self.get_blur_hints())

    def __setstate__(self, state):
        super().__setstate__(state)
        self.setBlurRadius(state["blur_radius"])
        self.set_blur_hints(*state["blur_hints"])

    def set_blur_hints(self, *hints: BlurHintStr):
        for item in hints:
            if item not in BLUR_HINTS:
                raise InvalidParamError(item, BLUR_HINTS)
        if hints:
            flags = helpers.merge_flags(hints, BLUR_HINTS)
        else:
            flags = QtWidgets.QGraphicsBlurEffect.BlurHint()
        self.setBlurHints(flags)

    def get_blur_hints(self) -> List[BlurHintStr]:
        return [k for k, v in BLUR_HINTS.items() if v & self.blurHints()]
